"""Types for the GSheets adapter."""
from enum import Enum


class SyncMode(Enum):
    """
    Different synchronization modes for GSheets.

    There are 3 different synchronization modes in which the adapter can
    operate. Each one has different tradeoffs.

    The ``BIDIRECTIONAL`` mode is useful for pushing small changes, and working
    interactively with the spreadsheets. All changes are pushed immediately
    to the sheet, and the sheet is fully downloaded before every ``UPDATE``
    or ``DELETE``.

    The ``BATCH`` mode is useful for one-off large changes, as the name
    suggests. Changes are push at once when the adapter closes, and the sheet
    is only downloaded once, before the first ``UPDATE`` or ``DELETE``.

    Finally, ``UNIDIRECTIONAL`` is a compromise between the other two modes.
    Changes are pushed immediately, since they're usually small. The sheet
    is downloaded in full only once, before the first ``UPDATE`` or ``DELETE``.
    """

    # all changes are pushed immediately, and the spreadsheet is downloaded
    # before every UPDATE/DELETE
    BIDIRECTIONAL = 1

    # all changes are pushed immediately, but the spreadsheet is
    # downloaded only once before the first UPDATE/DELETE
    UNIDIRECTIONAL = 2

    # all changes are pushed at once when the connection closes, and the
    # spreadsheet is downloaded before the first UPDATE/DELETE
    BATCH = 3



# pylint: disable=too-many-lines
"""
Tests for the Datasette adapter.
"""
import pytest
from pytest_mock import MockerFixture
from requests import Session
from requests_mock.mocker import Mocker

from shillelagh.adapters.api.datasette import (
    DatasetteAPI,
    get_field,
    is_datasette,
    is_known_domain,
)
from shillelagh.backends.apsw.db import connect
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import Float, Integer, ISODate, ISODateTime, String

from ...fakes import (
    datasette_columns_response,
    datasette_data_response_1,
    datasette_data_response_2,
    datasette_metadata_response,
    datasette_results,
)


def test_datasette(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test a simple query.
    """
    mocker.patch(
        "shillelagh.adapters.api.datasette.requests_cache.CachedSession",
        return_value=Session(),
    )

    columns_url = (
        "https://global-power-plants.datasettes.com/global-power-plants.json?"
        "sql=SELECT+*+FROM+%22global-power-plants%22+LIMIT+0"
    )
    requests_mock.get(columns_url, json=datasette_columns_response)
    metadata_url = (
        "https://global-power-plants.datasettes.com/global-power-plants.json?sql=SELECT+"
        "MAX%28%22country%22%29%2C+"
        "MAX%28%22country_long%22%29%2C+"
        "MAX%28%22name%22%29%2C+"
        "MAX%28%22gppd_idnr%22%29%2C+"
        "MAX%28%22capacity_mw%22%29%2C+"
        "MAX%28%22latitude%22%29%2C+"
        "MAX%28%22longitude%22%29%2C+"
        "MAX%28%22primary_fuel%22%29%2C+"
        "MAX%28%22other_fuel1%22%29%2C+"
        "MAX%28%22other_fuel2%22%29%2C+"
        "MAX%28%22other_fuel3%22%29%2C+"
        "MAX%28%22commissioning_year%22%29%2C+"
        "MAX%28%22owner%22%29%2C+"
        "MAX%28%22source%22%29%2C+"
        "MAX%28%22url%22%29%2C+"
        "MAX%28%22geolocation_source%22%29%2C+"
        "MAX%28%22wepp_id%22%29%2C+"
        "MAX%28%22year_of_capacity_data%22%29%2C+"
        "MAX%28%22generation_gwh_2013%22%29%2C+"
        "MAX%28%22generation_gwh_2014%22%29%2C+"
        "MAX%28%22generation_gwh_2015%22%29%2C+"
        "MAX%28%22generation_gwh_2016%22%29%2C+"
        "MAX%28%22generation_gwh_2017%22%29%2C+"
        "MAX%28%22generation_data_source%22%29%2C+"
        "MAX%28%22estimated_generation_gwh%22%29+"
        "FROM+%22global-power-plants%22+LIMIT+1"
    )
    requests_mock.get(metadata_url, json=datasette_metadata_response)
    data_url_1 = (
        "https://global-power-plants.datasettes.com/global-power-plants.json?"
        "sql=select+*+FROM+%22global-power-plants%22+WHERE+country+%3D+%27CAN%27+"
        "LIMIT+1001+OFFSET+0"
    )
    requests_mock.get(data_url_1, json=datasette_data_response_1)
    data_url_2 = (
        "https://global-power-plants.datasettes.com/global-power-plants.json?"
        "sql=select+*+FROM+%22global-power-plants%22+WHERE+country+%3D+%27CAN%27+"
        "LIMIT+1001+OFFSET+1000"
    )
    requests_mock.get(data_url_2, json=datasette_data_response_2)

    connection = connect(":memory:")
    cursor = connection.cursor()
    sql = """
        SELECT * FROM
        "https://global-power-plants.datasettes.com/global-power-plants/global-power-plants"
        WHERE country='CAN'
    """
    data = list(cursor.execute(sql))
    assert data == datasette_results


def test_datasette_no_data(mocker: MockerFixture) -> None:
    """
    Test result with no rows.
    """
    CachedSession = mocker.patch(  # pylint: disable=invalid-name
        "shillelagh.adapters.api.datasette.requests_cache.CachedSession",
    )
    CachedSession.return_value.get.return_value.json.return_value = {
        "columns": [],
        "rows": [],
    }

    with pytest.raises(ProgrammingError) as excinfo:
        DatasetteAPI("https://example.com", "database", "table")
    assert str(excinfo.value) == 'Table "table" has no data'


def test_get_metadata(requests_mock: Mocker) -> None:
    """
    Test ``get_metadata``.
    """
    requests_mock.get(
        "https://example.com/database.json?sql=SELECT+%2A+FROM+%22table%22+LIMIT+0",
        json={"columns": ["a", "b"]},
    )
    requests_mock.get(
        (
            "https://example.com/database.json?"
            "sql=SELECT+MAX%28%22a%22%29%2C+MAX%28%22b%22%29+FROM+%22table%22+LIMIT+1"
        ),
        json={"rows": [[1, 2], [3, 4]]},
    )
    requests_mock.get(
        "https://example.com/-/metadata.json",
        json={
            "databases": {"database": {"tables": {"table": {"foo": "bar"}}}},
        },
    )

    adapter = DatasetteAPI("https://example.com", "database", "table")
    assert adapter.get_metadata() == {"foo": "bar"}


def test_get_field() -> None:
    """
    Test ``get_field``.
    """
    assert isinstance(get_field(1), Integer)
    assert isinstance(get_field(1.1), Float)
    assert isinstance(get_field("test"), String)
    assert isinstance(get_field("2021-01-01"), ISODate)
    assert isinstance(get_field("2021-01-01 00:00:00"), ISODateTime)
    assert isinstance(get_field(None), String)


def test_is_known_domain() -> None:
    """
    Test ``is_known_domain``.
    """
    assert is_known_domain("latest.datasette.io")
    assert is_known_domain("san-francisco.datasettes.com")
    assert not is_known_domain("example.com")


def test_is_datasette(requests_mock: Mocker) -> None:
    """
    Test ``is_datasette``.
    """
    assert not is_datasette("https://example.com/")

    requests_mock.head(
        "https://example.com/-/versions.json",
        json={
            "python": {
                "version": "3.8.11",
                "full": "3.8.11 (default, Aug 17 2021, 15:56:41) \n[GCC 10.2.1 20210110]",
            },
            "datasette": {
                "version": "0.59a1",
                "note": "7e15422aacfa9e9735cb9f9beaa32250edbf4905",
            },
            "asgi": "3.0",
            "uvicorn": "0.15.0",
            "sqlite": {
                "version": "3.35.4",
                "fts_versions": ["FTS5", "FTS4", "FTS3"],
                "extensions": {"json1": None},
                "compile_options": [
                    "ALLOW_COVERING_INDEX_SCAN",
                    "COMPILER=gcc-4.8.2 20140120 (Red Hat 4.8.2-15)",
                    "ENABLE_FTS3",
                    "ENABLE_FTS3_PARENTHESIS",
                    "ENABLE_FTS4",
                    "ENABLE_FTS5",
                    "ENABLE_JSON1",
                    "ENABLE_LOAD_EXTENSION",
                    "ENABLE_RTREE",
                    "ENABLE_STAT4",
                    "ENABLE_UPDATE_DELETE_LIMIT",
                    "MAX_MMAP_SIZE=1099511627776",
                    "MAX_VARIABLE_NUMBER=250000",
                    "SOUNDEX",
                    "TEMP_STORE=3",
                    "THREADSAFE=1",
                    "USE_URI",
                ],
            },
            "pysqlite3": "0.4.6",
        },
    )
    assert is_datasette("https://example.com/database/table")


def test_datasette_error(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test error handling.
    """
    mocker.patch(
        "shillelagh.adapters.api.datasette.requests_cache.CachedSession",
        return_value=Session(),
    )

    columns_url = (
        "https://global-power-plants.datasettes.com/global-power-plants.json?"
        "sql=SELECT+*+FROM+%22global-power-plants%22+LIMIT+0"
    )
    requests_mock.get(columns_url, json=datasette_columns_response)
    metadata_url = (
        "https://global-power-plants.datasettes.com/global-power-plants.json?sql=SELECT+"
        "MAX%28%22country%22%29%2C+"
        "MAX%28%22country_long%22%29%2C+"
        "MAX%28%22name%22%29%2C+"
        "MAX%28%22gppd_idnr%22%29%2C+"
        "MAX%28%22capacity_mw%22%29%2C+"
        "MAX%28%22latitude%22%29%2C+"
        "MAX%28%22longitude%22%29%2C+"
        "MAX%28%22primary_fuel%22%29%2C+"
        "MAX%28%22other_fuel1%22%29%2C+"
        "MAX%28%22other_fuel2%22%29%2C+"
        "MAX%28%22other_fuel3%22%29%2C+"
        "MAX%28%22commissioning_year%22%29%2C+"
        "MAX%28%22owner%22%29%2C+"
        "MAX%28%22source%22%29%2C+"
        "MAX%28%22url%22%29%2C+"
        "MAX%28%22geolocation_source%22%29%2C+"
        "MAX%28%22wepp_id%22%29%2C+"
        "MAX%28%22year_of_capacity_data%22%29%2C+"
        "MAX%28%22generation_gwh_2013%22%29%2C+"
        "MAX%28%22generation_gwh_2014%22%29%2C+"
        "MAX%28%22generation_gwh_2015%22%29%2C+"
        "MAX%28%22generation_gwh_2016%22%29%2C+"
        "MAX%28%22generation_gwh_2017%22%29%2C+"
        "MAX%28%22generation_data_source%22%29%2C+"
        "MAX%28%22estimated_generation_gwh%22%29+"
        "FROM+%22global-power-plants%22+LIMIT+1"
    )
    requests_mock.get(metadata_url, json=datasette_metadata_response)
    data_url_1 = (
        "https://global-power-plants.datasettes.com/global-power-plants.json?"
        "sql=select+*+FROM+%22global-power-plants%22+WHERE+country+%3D+%27CAN%27+"
        "LIMIT+1001+OFFSET+0"
    )
    requests_mock.get(
        data_url_1,
        json={
            "ok": False,
            "error": "no such table: invalid",
            "status": 400,
            "title": "Invalid SQL",
        },
    )

    connection = connect(":memory:")
    cursor = connection.cursor()
    sql = """
        SELECT * FROM
        "https://global-power-plants.datasettes.com/global-power-plants/global-power-plants"
        WHERE country='CAN'
    """
    with pytest.raises(ProgrammingError) as excinfo:
        list(cursor.execute(sql))
    assert str(excinfo.value) == "Error (Invalid SQL): no such table: invalid"


@pytest.mark.integration_test
def test_integration(adapter_kwargs) -> None:
    """
    Test fetching data from the demo Datasette.
    """
    connection = connect(":memory:", adapter_kwargs=adapter_kwargs)
    cursor = connection.cursor()

    sql = """
        SELECT *
        FROM "https://global-power-plants.datasettes.com/global-power-plants/global-power-plants"
        WHERE wepp_id = ?
    """
    cursor.execute(sql, ("67644",))
    assert cursor.fetchall() == [
        (
            "USA",
            "United States of America",
            "145 Talmadge Solar",
            "USA0057458",
            3.8,
            40.5358,
            -74.3913,
            "Solar",
            None,
            None,
            None,
            2011.0,
            "Avidan Energy Solutions",
            "U.S. Energy Information Administration",
            "http://www.eia.gov/electricity/data/browser/",
            "U.S. Energy Information Administration",
            "67644",
            2017,
            5.0360000000000005,
            4.524,
            4.8020000000000005,
            5.051,
            4.819,
            "U.S. Energy Information Administration",
            None,
        ),
    ]



# pylint: disable=invalid-name
"""
Tests for shillelagh.console.
"""
import os.path
from io import StringIO

import yaml
from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_mock import MockerFixture

from shillelagh import console


def test_main(mocker: MockerFixture) -> None:
    """
    Test ``main``.
    """
    stdout = mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("shillelagh.console.PromptSession")

    PromptSession.return_value.prompt.side_effect = ["SELECT 1", "", EOFError()]
    console.main()
    result = stdout.getvalue()
    assert result == "  1\n---\n  1\nGoodBye!\n"


def test_exception(mocker: MockerFixture) -> None:
    """
    Test that exceptions are captured and printed.
    """
    stdout = mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("shillelagh.console.PromptSession")

    PromptSession.return_value.prompt.side_effect = ["SSELECT 1", EOFError()]
    console.main()
    result = stdout.getvalue()
    assert result == 'SQLError: near "SSELECT": syntax error\nGoodBye!\n'


def test_ctrl_c(mocker: MockerFixture) -> None:
    """
    Test that ``CTRL-C`` exists the REPL.
    """
    stdout = mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("shillelagh.console.PromptSession")

    PromptSession.return_value.prompt.side_effect = [
        KeyboardInterrupt(),
        "SELECT 1",
        EOFError(),
    ]
    console.main()
    result = stdout.getvalue()
    assert result == "  1\n---\n  1\nGoodBye!\n"


def test_configuration(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test loading the configuration file.
    """
    config = os.path.expanduser("~/.config/shillelagh/shillelagh.yaml")
    fs.create_file(config)
    with open(config, "w", encoding="utf-8") as fp:
        yaml.dump({"foo": {"bar": "baz"}}, fp)

    connect = mocker.patch("shillelagh.console.connect")
    mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("shillelagh.console.PromptSession")

    PromptSession.return_value.prompt.side_effect = [EOFError()]
    console.main()

    connect.assert_called_with(":memory:", adapter_kwargs={"foo": {"bar": "baz"}})


def test_no_configuration(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test no configuration file found.
    """
    config_dir = os.path.expanduser("~/.config/shillelagh/")
    fs.create_dir(config_dir)

    connect = mocker.patch("shillelagh.console.connect")
    mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("shillelagh.console.PromptSession")

    PromptSession.return_value.prompt.side_effect = [EOFError()]
    console.main()

    connect.assert_called_with(":memory:", adapter_kwargs={})


def test_configuration_invalid(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test that an exception is raised if the configuration is invalid.
    """
    config = os.path.expanduser("~/.config/shillelagh/shillelagh.yaml")
    fs.create_file(config)
    with open(config, "w", encoding="utf-8") as fp:
        fp.write("foo: *")

    _logger = mocker.patch("shillelagh.console._logger")
    mocker.patch("sys.stdout", new_callable=StringIO)
    PromptSession = mocker.patch("shillelagh.console.PromptSession")

    PromptSession.return_value.prompt.side_effect = [EOFError()]
    console.main()

    _logger.exception.assert_called_with("Unable to load configuration file")



# pylint: disable=abstract-method
"""
A "safe" Shillelagh dialect.

When this dialect is used only adapters marked as safe and explicitly
listed are loaded.
"""
from typing import Any, Dict, List, Optional, Tuple

from sqlalchemy.engine.url import URL

from shillelagh.backends.apsw.dialects.base import APSWDialect


class APSWSafeDialect(APSWDialect):

    """
    A "safe" Shillelagh dialect.

    This dialect can be used with the `shillelagh+safe://` URI:

        >>> from sqlalchemy.engine import create_engine
        >>> engine = create_engine("shillelagh+safe://", adapters=["socrata"])

    The dialect only loads the adapters explicitly listed ("socrata", in the
    example above), and only if they're marked as safe.

    """

    def __init__(
        self,
        adapters: Optional[List[str]] = None,
        adapter_kwargs: Optional[Dict[str, Dict[str, Any]]] = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self._adapters = adapters
        self._adapter_kwargs = adapter_kwargs or {}
        self._safe = True

    def create_connect_args(
        self,
        url: URL,
    ) -> Tuple[Tuple[()], Dict[str, Any]]:
        return (), {
            "path": ":memory:",
            "adapters": self._adapters,
            "adapter_kwargs": self._adapter_kwargs,
            "safe": True,
            "isolation_level": self.isolation_level,
        }



"""
Test the date/time pattern handling (parsing and formatting).
"""
# pylint: disable=protected-access
from datetime import date, datetime, time, timedelta

import pytest

from shillelagh.adapters.api.gsheets.parsing.base import LITERAL
from shillelagh.adapters.api.gsheets.parsing.date import (
    AMPM,
    AP,
    DD,
    DDD,
    MM,
    MMM,
    MMMM,
    MMMMM,
    SS,
    YY,
    YYYY,
    ZERO,
    D,
    DDDDPlus,
    H,
    HHPlus,
    HPlusDuration,
    M,
    Meridiem,
    MPlusDuration,
    S,
    SPlusDuration,
    format_date_time_pattern,
    parse_date_time_pattern,
    tokenize,
)

classes = [
    H,
    HHPlus,
    M,
    MM,
    MMM,
    MMMM,
    MMMMM,
    S,
    SS,
    HPlusDuration,
    MPlusDuration,
    SPlusDuration,
    D,
    DD,
    DDD,
    DDDDPlus,
    YY,
    YYYY,
    AP,
    AMPM,
    ZERO,
    LITERAL,
]


def test_implementation():
    """
    Test the examples from the reference.

    https://developers.google.com/sheets/api/guides/formats?hl=en
    """
    timestamp = datetime(2016, 4, 5, 16, 8, 53, 528000)

    assert format_date_time_pattern(timestamp, "h:mm:ss.00 a/p") == "4:08:53.53 p"
    assert parse_date_time_pattern("4:08:53.53 p", "h:mm:ss.00 a/p", time) == time(
        16,
        8,
        53,
        530000,
    )

    assert format_date_time_pattern(timestamp, 'hh:mm A/P".M."') == "04:08 P.M."
    assert parse_date_time_pattern("04:08 P.M.", 'hh:mm A/P".M."', time) == time(16, 8)

    assert format_date_time_pattern(timestamp, "yyyy-mm-dd") == "2016-04-05"
    assert parse_date_time_pattern("2016-04-05", "yyyy-mm-dd", date) == date(2016, 4, 5)

    # unsupported parse
    assert (
        format_date_time_pattern(timestamp, r"mmmm d \[dddd\]") == "April 5 [Tuesday]"
    )
    assert format_date_time_pattern(timestamp, "h PM, ddd mmm dd") == "4 PM, Tue Apr 05"

    assert (
        format_date_time_pattern(timestamp, "dddd, m/d/yy at h:mm")
        == "Tuesday, 4/5/16 at 16:08"
    )
    assert (
        parse_date_time_pattern(
            "Tuesday, 4/5/16 at 16:08",
            "dddd, m/d/yy at h:mm",
            datetime,
        )
        == datetime(2016, 4, 5, 16, 8)
    )

    duration = timedelta(hours=3, minutes=13, seconds=41, microseconds=255000)

    assert format_date_time_pattern(duration, "[hh]:[mm]:[ss].000") == "03:13:41.255"
    assert (
        parse_date_time_pattern("03:13:41.255", "[hh]:[mm]:[ss].000", timedelta)
        == duration
    )

    assert format_date_time_pattern(duration, "[mmmm]:[ss].000") == "0193:41.255"
    assert (
        parse_date_time_pattern("0193:41.255", "[mmmm]:[ss].000", timedelta) == duration
    )


def test_token():
    """
    General tests for tokens.
    """
    assert H("h") == H("h")
    assert H("h") != M("m")


def test_h_token():
    """
    Test the h token.
    """
    token = H("h")

    assert H.match("h:mm:ss")
    assert not H.match("hh:mm:ss")
    assert not H.match("s")

    assert H.consume("h:mm:ss") == (token, ":mm:ss")
    with pytest.raises(Exception) as excinfo:
        H.consume("hh:mm:ss")
    assert str(excinfo.value) == "Token could not find match"

    tokens = list(tokenize("h:mm:ss", classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "13"
    assert token.format(datetime(2021, 11, 12, 3, 14, 15, 16), tokens) == "3"
    tokens = list(tokenize("h:mm:ss a/p", classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "1"
    tokens = list(tokenize('h:mm:ss "PM"', classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "1"

    tokens = list(tokenize("h:mm:ss", classes))
    assert token.parse("123", tokens) == ({"hour": 12}, "3")
    assert token.parse("303", tokens) == ({"hour": 3}, "03")
    with pytest.raises(Exception) as excinfo:
        token.parse("invalid", tokens)
    assert str(excinfo.value) == "Cannot parse value: invalid"


def test_hhplus_token():
    """
    Test the hhh+ token.
    """
    token = HHPlus("hhh")

    assert HHPlus.match("hhh:mm:ss")
    assert HHPlus.match("hhhh:mm:ss")
    assert not HHPlus.match("h:mm:ss")

    tokens = list(tokenize("hhh:mm:ss", classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "13"
    assert token.format(datetime(2021, 11, 12, 3, 14, 15, 16), tokens) == "03"
    tokens = list(tokenize("hhh:mm:ss a/p", classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "01"

    tokens = list(tokenize("hhh:mm:ss", classes))
    assert token.parse("123", tokens) == ({"hour": 12}, "3")
    assert token.parse("303", tokens) == ({"hour": 30}, "3")


def test_m_token():
    """
    Test the m token.
    """
    token = M("m")

    assert M.match("m/d/y")
    assert not M.match("mm/dd/yyyy")
    assert not M.match("M/d/y")

    assert M.consume("m/d/y") == (token, "/d/y")

    tokens = list(tokenize("m/d/y", classes))
    token = tokens[0]
    assert token._is_minute(tokens) is False
    tokens = list(tokenize("h//m", classes))
    token = tokens[2]
    assert token._is_minute(tokens) is True
    tokens = list(tokenize("m//s", classes))
    token = tokens[0]
    assert token._is_minute(tokens) is True

    tokens = list(tokenize("m/d/y", classes))
    with pytest.raises(Exception) as excinfo:
        token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens)
    assert str(excinfo.value) == "Token is not present in list of tokens"
    token = tokens[0]
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "11"
    tokens = list(tokenize("h:m:s", classes))
    token = tokens[2]
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "14"
    tokens = list(tokenize("m/d/y", classes))
    token = tokens[0]
    with pytest.raises(Exception) as excinfo:
        token.format(time(13, 14, 15, 16), tokens)
    assert str(excinfo.value) == "Cannot format value: 13:14:15.000016"

    tokens = list(tokenize("m/d/y", classes))
    token = tokens[0]
    assert token.parse("123", tokens) == ({"month": 12}, "3")
    assert token.parse("303", tokens) == ({"month": 3}, "03")
    tokens = list(tokenize("h:m:s", classes))
    token = tokens[2]
    assert token.parse("14:15", tokens) == ({"minute": 14}, ":15")
    with pytest.raises(Exception) as excinfo:
        token.parse("invalid", tokens)
    assert str(excinfo.value) == "Cannot parse value: invalid"


def test_mm_token():
    """
    Test the mm token.
    """
    token = MM("mm")

    assert MM.match("mm/dd/yyy")
    assert not MM.match("mmm/dd/yyy")

    assert MM.consume("mm/dd/yyyy") == (token, "/dd/yyyy")

    tokens = list(tokenize("mm/dd/yyy", classes))
    token = tokens[0]
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "11"
    assert token.parse("123", tokens) == ({"month": 12}, "3")


def test_mmm_token():
    """
    Test the mmm token.
    """
    token = MMM("mmm")

    assert MMM.match("mmm/dd/yyy")
    assert not MMM.match("mm/dd/yyy")

    assert MMM.consume("mmm/dd/yyyy") == (token, "/dd/yyyy")

    tokens = list(tokenize("mmm/dd/yyy", classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "Nov"
    assert token.parse("Mar 1st", tokens) == ({"month": 3}, " 1st")


def test_mmmm_token():
    """
    Test the mmm token.
    """
    token = MMMM("mmmm")

    assert MMMM.match("mmmm/dd/yyy")
    assert not MMMM.match("mmm/dd/yyy")

    assert MMMM.consume("mmmm/dd/yyyy") == (token, "/dd/yyyy")

    tokens = list(tokenize("mmmm/dd/yyy", classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "November"
    assert token.parse("March 1st", tokens) == ({"month": 3}, " 1st")


def test_mmmmm_token():
    """
    Test the mmmmm token.
    """
    token = MMMMM("mmmmm")

    assert MMMMM.match("mmmmm/dd/yyy")
    assert not MMMMM.match("mmm/dd/yyy")

    assert MMMMM.consume("mmmmm/dd/yyyy") == (token, "/dd/yyyy")

    tokens = list(tokenize("mmmmm/dd/yyy", classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "N"
    assert token.parse("F 1st", token) == ({"month": 2}, " 1st")
    with pytest.raises(Exception) as excinfo:
        token.parse("Z 1st", tokens)
    assert str(excinfo.value) == "Unable to find month letter: Z"
    with pytest.raises(Exception) as excinfo:
        token.parse("M 1st", tokens)
    assert str(excinfo.value) == "Unable to parse month letter unambiguously: M"


def test_s_token():
    """
    Test the s token.
    """
    token = S("s")

    assert S.match("s.00")
    assert not S.match("ss.00")

    assert S.consume("s.00") == (token, ".00")

    tokens = list(tokenize("h:m:s", classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 5, 16), tokens) == "5"
    assert token.format(datetime(2021, 11, 12, 13, 14, 5, 16), tokens) == "5"
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "15"
    assert token.parse("59.123", tokens) == ({"second": 59}, ".123")
    assert token.parse("60.123", tokens) == ({"second": 60}, ".123")
    assert token.parse("61.123", tokens) == ({"second": 61}, ".123")
    assert token.parse("62.123", tokens) == ({"second": 6}, "2.123")
    with pytest.raises(Exception) as excinfo:
        token.parse("invalid", tokens)
    assert str(excinfo.value) == "Cannot parse value: invalid"


def test_ss_token():
    """
    Test the ss token.
    """
    token = SS("ss")

    assert SS.match("ss.00")
    assert not SS.match("s.00")

    assert SS.consume("ss.00") == (token, ".00")

    tokens = list(tokenize("hh:mm:ss", classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "15"
    assert token.format(datetime(2021, 11, 12, 13, 14, 5, 16), tokens) == "05"
    assert token.parse("59.123", tokens) == ({"second": 59}, ".123")


def test_hplusduration_token():
    """
    Test the [h+] token.
    """
    token = HPlusDuration("[hh]")

    assert HPlusDuration.match("[h]")
    assert HPlusDuration.match("[hh]")
    assert not HPlusDuration.match("hh")

    assert HPlusDuration.consume("[hh]:[mm]:[ss].000") == (token, ":[mm]:[ss].000")

    tokens = list(tokenize("[hh]:[mm]:[ss].000", classes))
    assert (
        token.format(
            timedelta(days=1, hours=2, minutes=3, seconds=4, microseconds=500000),
            tokens,
        )
        == "26"
    )
    assert (
        token.format(
            timedelta(hours=2, minutes=3, seconds=4, microseconds=500000),
            tokens,
        )
        == "02"
    )
    assert token.parse("26:03:04.500", "[hh]:[mm]:[ss].000") == (
        {"hours": 26},
        ":03:04.500",
    )


def test_mplusduration_token():
    """
    Test the [m+] token.
    """
    token = MPlusDuration("[mm]")

    assert MPlusDuration.match("[m]")
    assert MPlusDuration.match("[mm]")
    assert not MPlusDuration.match("mm")

    assert MPlusDuration.consume("[mm]:[ss].000") == (token, ":[ss].000")

    tokens = list(tokenize("[hh]:[mm]:[ss].000", classes))
    assert (
        token.format(
            timedelta(days=1, hours=2, minutes=3, seconds=4, microseconds=500000),
            tokens,
        )
        == "03"
    )

    token = MPlusDuration("[mmmm]")
    tokens = list(tokenize("[mmmm]:[ss].000", classes))
    assert (
        token.format(
            timedelta(hours=2, minutes=3, seconds=4, microseconds=500000),
            tokens,
        )
        == "0123"
    )

    token = MPlusDuration("[mm]")
    tokens = list(tokenize("[hh]:[mm]:[ss].000", classes))
    assert token.parse("03:04.500", "[hh]:[mm]:[ss].000") == (
        {"minutes": 3},
        ":04.500",
    )


def test_splusduration_token():
    """
    Test the [s+] token.
    """
    token = SPlusDuration("[ss]")

    assert SPlusDuration.match("[s]")
    assert SPlusDuration.match("[ss]")
    assert not SPlusDuration.match("ss")

    assert SPlusDuration.consume("[ss].000") == (token, ".000")

    tokens = list(tokenize("[hh]:[mm]:[ss].000", classes))
    assert (
        token.format(
            timedelta(days=1, hours=2, minutes=3, seconds=4, microseconds=500000),
            tokens,
        )
        == "04"
    )

    tokens = list(tokenize("[ss].000", classes))
    assert (
        token.format(
            timedelta(hours=2, minutes=3, seconds=4, microseconds=500000),
            tokens,
        )
        == "7384"
    )

    tokens = list(tokenize("[hh]:[mm]:[ss].000", classes))
    assert token.parse("04.500", "[hh]:[mm]:[ss].000") == (
        {"seconds": 4},
        ".500",
    )


def test_d_token():
    """
    Test the d token.
    """
    token = D("d")

    assert D.match("d")
    assert not D.match("dd")

    assert D.consume("d/m/y") == (token, "/m/y")

    tokens = list(tokenize("d/m/y", classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "12"
    assert token.format(datetime(2021, 11, 2, 13, 14, 15, 16), tokens) == "2"
    assert token.parse("12/11/21", tokens) == ({"day": 12}, "/11/21")
    with pytest.raises(Exception) as excinfo:
        token.parse("invalid", tokens)
    assert str(excinfo.value) == "Cannot parse value: invalid"


def test_dd_token():
    """
    Test the dd token.
    """
    token = DD("dd")

    assert DD.match("dd")
    assert not DD.match("d")
    assert not DD.match("ddd")

    assert DD.consume("dd/mm/yyyy") == (token, "/mm/yyyy")

    tokens = list(tokenize("dd/mm/yyyy", classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "12"
    assert token.format(datetime(2021, 11, 2, 13, 14, 15, 16), tokens) == "02"
    assert token.parse("12/11/2021", tokens) == ({"day": 12}, "/11/2021")


def test_ddd_token():
    """
    Test the ddd token.
    """
    token = DDD("ddd")

    assert DDD.match("ddd")
    assert not DDD.match("dd")
    assert not DDD.match("dddd")

    assert DDD.consume("ddd, dd/mm/yyyy") == (token, ", dd/mm/yyyy")

    tokens = list(tokenize("ddd, dd/mm/yyyy", classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "Fri"
    assert token.parse("Fri, 12/11/2021", tokens) == ({"weekday": 0}, ", 12/11/2021")


def test_ddddplus_token():
    """
    Test the dddd+ token.
    """
    token = DDDDPlus("dddd")

    assert DDDDPlus.match("dddd")
    assert DDDDPlus.match("ddddd")
    assert not DDDDPlus.match("ddd")

    assert DDDDPlus.consume("dddd, dd/mm/yyyy") == (token, ", dd/mm/yyyy")

    tokens = list(tokenize("dddd, dd/mm/yyyy", classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "Friday"
    assert token.parse("Friday, 12/11/2021", tokens) == ({"weekday": 0}, ", 12/11/2021")


def test_yy_token():
    """
    Test the yy token.
    """
    token = YY("yy")

    assert YY.match("y")
    assert YY.match("yy")
    assert not YY.match("yyy")

    assert YY.consume("yy/mm/dd") == (token, "/mm/dd")

    tokens = list(tokenize("yy/mm/dd", classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "21"
    assert token.parse("21/11/12", tokens) == ({"year": 2021}, "/11/12")


def test_yyyy_token():
    """
    Test the yyyy token.
    """
    token = YYYY("yyyy")

    assert YYYY.match("yyy")
    assert YYYY.match("yyyy")
    assert not YYYY.match("yy")

    assert YYYY.consume("yyyy/mm/dd") == (token, "/mm/dd")

    tokens = list(tokenize("yyyy/mm/dd", classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "2021"
    assert token.parse("2021/11/12", tokens) == ({"year": 2021}, "/11/12")


def test_zero_token():
    """
    Test the 00 token.
    """
    token = ZERO("00")

    assert ZERO.match("0")
    assert ZERO.match("00")
    assert ZERO.match("000")
    assert not ZERO.match("0000")

    assert ZERO.consume("00") == (token, "")

    tokens = list(tokenize("hh:mm:ss.000", classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 167000), tokens) == "17"

    token = ZERO("000")
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 167000), tokens) == "167"

    assert token.parse("123", tokens) == ({"microsecond": 123000}, "")


def test_ap_token():
    """
    Test the a/p token.
    """
    token = AP("a/p")

    assert AP.match("a/p")
    assert AP.match("A/P")
    assert not AP.match("AM")

    assert AP.consume("a/p hh:mm") == (token, " hh:mm")

    tokens = list(tokenize("a/p hh:mm", classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "p"
    assert token.format(datetime(2021, 11, 12, 1, 14, 15, 16), tokens) == "a"

    tokens = list(tokenize("A/P hh:mm", classes))
    token = AP("A/P")
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "P"

    assert token.parse("P 09:00", tokens) == ({"meridiem": Meridiem.PM}, " 09:00")
    assert token.parse("A 09:00", tokens) == ({"meridiem": Meridiem.AM}, " 09:00")


def test_ampm_token():
    """
    Test the am/pm token.
    """
    token = AMPM("am/pm")

    assert AMPM.match("am/pm")
    assert not AMPM.match("AM/PM")
    assert not AMPM.match("AM")

    assert AMPM.consume("am/pm hh:mm") == (token, " hh:mm")

    tokens = list(tokenize("am/pm hh:mm", classes))
    assert token.format(datetime(2021, 11, 12, 13, 14, 15, 16), tokens) == "PM"
    assert token.format(datetime(2021, 11, 12, 1, 14, 15, 16), tokens) == "AM"

    assert token.parse("PM 09:00", tokens) == ({"meridiem": Meridiem.PM}, " 09:00")
    assert token.parse("AM 09:00", tokens) == ({"meridiem": Meridiem.AM}, " 09:00")


def test_parse_date_time_pattern():
    """
    Test the parse_date_time_pattern function.
    """
    parsed = parse_date_time_pattern(
        "2021/11/12 13:14:15.167",
        "yyyy/mm/dd hh:mm:ss.000",
        datetime,
    )
    assert parsed == datetime(2021, 11, 12, 13, 14, 15, 167000)

    parsed = parse_date_time_pattern(
        "2021/11/12 01:14:15.167 PM",
        "yyyy/mm/dd hh:mm:ss.000 am/pm",
        datetime,
    )
    assert parsed == datetime(2021, 11, 12, 13, 14, 15, 167000)

    # test that weekday is ignored
    parsed = parse_date_time_pattern(
        "2021/11/12 Fri, 01:14:15.167 PM",
        "yyyy/mm/dd ddd, hh:mm:ss.000 am/pm",
        datetime,
    )
    assert parsed == datetime(2021, 11, 12, 13, 14, 15, 167000)

    parsed = parse_date_time_pattern("60.123", "[ss].000", timedelta)
    assert parsed == timedelta(seconds=60, microseconds=123000)

    with pytest.raises(Exception) as excinfo:
        parse_date_time_pattern("60.123", "[ss].000", datetime)
    assert str(excinfo.value) == "Unsupported format"


def test_format_date_time_pattern():
    """
    Test the format_date_time_pattern function.
    """
    assert (
        format_date_time_pattern(
            datetime(2021, 11, 12, 13, 14, 15, 16),
            "yyyy/mm/dd hh/mm/ss.000",
        )
        == "2021/11/12 13/14/15.000"
    )


def test_parse_date_time_pattern_with_quotes():
    """
    Test parsing a timestamp with quotes.
    """
    parsed = parse_date_time_pattern(
        "1/1/2021",
        'm"/"d"/"yyyy',
        date,
    )
    assert parsed == date(2021, 1, 1)


def test_parse_date_time_with_meridiem():
    """
    Test parsing a timestamp with AM/PM in the hour.
    """
    parsed = parse_date_time_pattern(
        "12:34:56 AM",
        "h:mm:ss am/pm",
        time,
    )
    assert parsed == time(0, 34, 56)

    parsed = parse_date_time_pattern(
        "12:34:56 PM",
        "h:mm:ss am/pm",
        time,
    )
    assert parsed == time(12, 34, 56)


def test_parse_date_time_without_meridiem():
    """
    Test parsing a timestamp without AM/PM in the hour.
    """
    parsed = parse_date_time_pattern(
        "12/31/2020 12:34:56",
        "m/d/yyyy h:mm:ss",
        datetime,
    )
    assert parsed == datetime(2020, 12, 31, 12, 34, 56)


def test_format_date_time_with_meridiem():
    """
    Test formatting a timestamp with AM/pM in the hour.
    """
    assert (
        format_date_time_pattern(
            time(12, 34, 56),
            "h:mm:ss am/pm",
        )
        == "12:34:56 PM"
    )



"""
Custom fields for the GSheets adapter.
"""
import datetime
from distutils.util import strtobool
from typing import Any, List, Optional, Type

from shillelagh.adapters.api.gsheets.parsing.date import (
    format_date_time_pattern,
    parse_date_time_pattern,
)
from shillelagh.adapters.api.gsheets.parsing.number import (
    format_number_pattern,
    parse_number_pattern,
)
from shillelagh.fields import External, Field, Internal, Order
from shillelagh.filters import Filter

# timestamp format used in SQL queries
DATETIME_SQL_QUOTE = "%Y-%m-%d %H:%M:%S"
DATE_SQL_QUOTE = "%Y-%m-%d"
TIME_SQL_QUOTE = "%H:%M:%S"


class GSheetsField(Field[Internal, External]):
    """
    A base class for GSheets fields.
    """

    # the default formats for date and datetime do not follow the same syntax
    # as user-configured formats (I suspect it uses ICU instead)
    pattern_substitutions = {
        "M/d/yyyy H:mm:ss": "m/d/yyyy h:mm:ss",
        "M/d/yyyy": "m/d/yyyy",
    }

    def __init__(  # pylint: disable=too-many-arguments
        self,
        filters: Optional[List[Type[Filter]]] = None,
        order: Order = Order.NONE,
        exact: bool = False,
        pattern: Optional[str] = None,
        timezone: Optional[datetime.tzinfo] = None,
    ):
        super().__init__(filters, order, exact)
        self.pattern: Optional[str] = (
            self.pattern_substitutions[pattern]
            if pattern in self.pattern_substitutions
            else pattern
        )
        self.timezone = timezone

    def __eq__(self, other: Any) -> bool:
        if other.__class__ != self.__class__:
            return NotImplemented

        return bool(
            self.filters == other.filters
            and self.order == other.order
            and self.exact == other.exact
            and self.pattern == other.pattern
            and self.timezone == other.timezone,
        )


class GSheetsDateTime(GSheetsField[str, datetime.datetime]):
    """
    A GSheets timestamp.

    The Google Chart API returns timestamps as a string encoded using an ICU
    pattern. The default format is "M/d/yyyy H:mm:ss", and values look like
    this:

        "9/1/2018 0:00:00"

    There are no timezones; instead, there is a global timezone for the whole
    spreadsheet. The timezone can only be read if the user has set their
    credentials, since the Google Sheets API used to read metadata about the
    sheet requires authentication.

    When the timezone is present and read successfully all timestamps are
    converted to it, both when fetching data as well as when inserting rows.
    """

    type = "TIMESTAMP"
    db_api_type = "DATETIME"

    def parse(self, value: Optional[str]) -> Optional[datetime.datetime]:
        # Google Chart API returns ``None`` for a NULL cell, while the Google
        # Sheets API returns an empty string
        if self.pattern is None or value is None or value == "":
            return None

        timestamp = parse_date_time_pattern(value, self.pattern, datetime.datetime)

        # Set the timestamp to the spreadsheet timezone, if any.
        timestamp = timestamp.replace(tzinfo=self.timezone)

        return timestamp

    def format(self, value: Optional[datetime.datetime]) -> str:
        # This method is used only when inserting or updating rows, so we
        # encode NULLs as an empty string to match the Google Sheets API.
        if self.pattern is None or value is None:
            return ""

        # Google Sheets does not support timezones in datetime values, so we
        # convert all timestamps to the sheet timezone
        if self.timezone:
            value = value.astimezone(self.timezone)

        return format_date_time_pattern(value, self.pattern)

    def quote(self, value: Optional[str]) -> str:
        if self.pattern is None or value == "" or value is None:
            return "null"

        # On SQL queries the timestamp should be prefix by "datetime"
        value = parse_date_time_pattern(
            value,
            self.pattern,
            datetime.datetime,
        ).strftime(DATETIME_SQL_QUOTE)
        return f"datetime '{value}'"


class GSheetsDate(GSheetsField[str, datetime.date]):
    """
    A GSheets date.

    The Google Chart API returns dates as a string encoded using an ICU
    pattern. The default format is "M/d/yyyy", and values look like
    this:

        "9/1/2018"

    """

    type = "DATE"
    db_api_type = "DATETIME"

    def parse(self, value: Optional[str]) -> Optional[datetime.date]:
        # Google Chart API returns ``None`` for a NULL cell, while the Google
        # Sheets API returns an empty string
        if self.pattern is None or value is None or value == "":
            return None

        return parse_date_time_pattern(value, self.pattern, datetime.date)

    def format(self, value: Optional[datetime.date]) -> str:
        if self.pattern is None or value is None:
            return ""

        return format_date_time_pattern(value, self.pattern)

    def quote(self, value: Optional[str]) -> str:
        if self.pattern is None or value == "" or value is None:
            return "null"

        # On SQL queries the timestamp should be prefix by "date"
        value = parse_date_time_pattern(value, self.pattern, datetime.date).strftime(
            DATE_SQL_QUOTE,
        )
        return f"date '{value}'"


class GSheetsTime(GSheetsField[str, datetime.time]):
    """
    A GSheets time.

    The Google Chart API returns times as a string encoded using an ICU
    pattern. The default format is "h:mm:ss am/pm", and values look like
    this:

        "5:00:00 PM"

    """

    type = "TIME"
    db_api_type = "DATETIME"

    def parse(self, value: Optional[str]) -> Optional[datetime.time]:
        """
        Parse time of day as returned from the Google Chart API.
        """
        # Google Chart API returns ``None`` for a NULL cell, while the Google
        # Sheets API returns an empty string
        if self.pattern is None or value is None or value == "":
            return None

        return parse_date_time_pattern(value, self.pattern, datetime.time)

    def format(self, value: Optional[datetime.time]) -> str:
        if self.pattern is None or value is None:
            return ""

        return format_date_time_pattern(value, self.pattern)

    def quote(self, value: Optional[str]) -> str:
        if self.pattern is None or value == "" or value is None:
            return "null"

        # On SQL queries the timestamp should be prefix by "timeofday"
        value = parse_date_time_pattern(value, self.pattern, datetime.time).strftime(
            TIME_SQL_QUOTE,
        )
        return f"timeofday '{value}'"


class GSheetsBoolean(GSheetsField[str, bool]):
    """
    A GSheets boolean.

    Booleans in the Google Chart API are return as a string, either "TRUE"
    of "FALSE".
    """

    type = "BOOLEAN"
    db_api_type = "NUMBER"

    def parse(self, value: Optional[str]) -> Optional[bool]:
        # Google Chart API returns ``None`` for a NULL cell, while the Google
        # Sheets API returns an empty string
        if value is None or value == "":
            return None

        return bool(strtobool(value))

    def format(self, value: Optional[bool]) -> str:
        if value is None:
            return ""

        return "TRUE" if value else "FALSE"

    def quote(self, value: Optional[str]) -> str:
        if value == "" or value is None:
            return "null"
        return value.lower()


class GSheetsNumber(GSheetsField[str, float]):
    """
    A GSheets number.

    The Google Chart/Sheets APIs return "numbers" only, encoded as strings.
    """

    type = "REAL"
    db_api_type = "NUMBER"

    def parse(self, value: Optional[str]) -> Optional[float]:
        if value is None or value == "":
            return None

        if self.pattern is None or self.pattern == "General":
            try:
                return int(value)
            except ValueError:
                return float(value)

        return parse_number_pattern(value, self.pattern)

    def format(self, value: Optional[float]) -> str:
        if value is None:
            return ""

        if self.pattern is None or self.pattern == "General":
            return str(value)

        return format_number_pattern(value, self.pattern)

    def quote(self, value: Optional[str]) -> str:
        if value == "" or value is None:
            return "null"

        return str(value)


class GSheetsString(GSheetsField[str, str]):
    """
    A GSheets string.
    """

    type = "TEXT"
    db_api_type = "STRING"

    def parse(self, value: Optional[str]) -> Optional[str]:
        return None if value == "" else value

    def format(self, value: Optional[str]) -> str:
        return "" if value is None else value

    def quote(self, value: Optional[str]) -> str:
        if value == "" or value is None:
            return "null"

        return f"'{value}'"



"""
An adapter to WeatherAPI (https://www.weatherapi.com/).
"""
import logging
import urllib.parse
from datetime import date, datetime, timedelta, timezone
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union, cast

import dateutil.parser
import dateutil.tz
import requests_cache

from shillelagh.adapters.base import Adapter
from shillelagh.exceptions import ImpossibleFilterError
from shillelagh.fields import DateTime, Float, IntBoolean, Integer, Order, String
from shillelagh.filters import Filter, Impossible, Operator, Range
from shillelagh.typing import RequestedOrder, Row

_logger = logging.getLogger(__name__)

INITIAL_COST = 0
FETCHING_COST = 1000


def combine_time_filters(bounds: Dict[str, Filter]) -> Range:
    """
    Combine both time filters together.

    The adapter has two time columns that can be used to filter the data, "time" as
    a timestamp and "time_epoch" as a float. We convert the latter to a timestamp and
    combine the two filters into a single ``Range``.
    """
    time_range = bounds.get("time", Range())
    time_epoch_range = bounds.get("time_epoch", Range())

    if isinstance(time_range, Impossible) or isinstance(time_epoch_range, Impossible):
        raise ImpossibleFilterError()

    if not isinstance(time_range, Range) or not isinstance(time_epoch_range, Range):
        raise Exception("Invalid filter")

    # convert time_epoch range to datetime so we can combine it
    # with the time range
    time_epoch_range.start = (
        datetime.fromtimestamp(time_epoch_range.start, tz=timezone.utc)
        if time_epoch_range.start is not None
        else None
    )
    time_epoch_range.end = (
        datetime.fromtimestamp(time_epoch_range.end, tz=timezone.utc)
        if time_epoch_range.end is not None
        else None
    )

    # combine time ranges together and check if the result is a valid range
    time_range += time_epoch_range
    if isinstance(time_range, Impossible):
        raise ImpossibleFilterError()

    return cast(Range, time_range)


class WeatherAPI(Adapter):

    """
    An adapter for WeatherAPI (https://www.weatherapi.com/).

    The adapter expects an URL like::

        https://api.weatherapi.com/v1/history.json?key=$key&q=$location

    Where ``$key`` is an API key (available for free), and ``$location`` is a
    freeform value that can be a US Zipcode, UK Postcode, Canada Postalcode,
    IP address, Latitude/Longitude (decimal degree) or city name.
    """

    safe = True

    # These two columns can be used to filter the results from the API. We
    # define them as inexact since we will retrieve data for the whole day,
    # even if specific hours are requested. The post-filtering will be done
    # by the backend.
    time = DateTime(filters=[Range], order=Order.ASCENDING, exact=False)
    time_epoch = Float(filters=[Range], order=Order.ASCENDING, exact=False)

    temp_c = Float()
    temp_f = Float()
    is_day = IntBoolean()
    wind_mph = Float()
    wind_kph = Float()
    wind_degree = Integer()
    wind_dir = String()
    pressure_mb = Float()
    pressure_in = Float()
    precip_mm = Float()
    precip_in = Float()
    humidity = Integer()
    cloud = Integer()
    feelslike_c = Float()
    feelslike_f = Float()
    windchill_c = Float()
    windchill_f = Float()
    heatindex_c = Float()
    heatindex_f = Float()
    dewpoint_c = Float()
    dewpoint_f = Float()
    will_it_rain = IntBoolean()
    chance_of_rain = String()
    will_it_snow = IntBoolean()
    chance_of_snow = String()
    vis_km = Float()
    vis_miles = Float()
    gust_mph = Float()
    gust_kph = Float()

    @staticmethod
    def supports(uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        """https://api.weatherapi.com/v1/history.json?key=XXX&q=94158"""
        parsed = urllib.parse.urlparse(uri)
        query_string = urllib.parse.parse_qs(parsed.query)
        return (
            parsed.netloc == "api.weatherapi.com"
            and parsed.path == "/v1/history.json"
            and "q" in query_string
            and ("key" in query_string or "api_key" in kwargs)
        )

    @staticmethod
    def parse_uri(uri: str) -> Union[Tuple[str], Tuple[str, str]]:
        parsed = urllib.parse.urlparse(uri)
        query_string = urllib.parse.parse_qs(parsed.query)
        location = query_string["q"][0]

        # key can be passed in the URL or via connection arguments
        if "key" in query_string:
            return (location, query_string["key"][0])
        return (location,)

    def __init__(self, location: str, api_key: str, window: int = 7):
        super().__init__()

        self.location = location
        self.api_key = api_key
        self.window = window

        # use a cache, since the adapter does a lot of similar API requests,
        # and the data should rarely (never?) change
        self._session = requests_cache.CachedSession(
            cache_name="weatherapi_cache",
            backend="sqlite",
            expire_after=180,
        )

    def get_cost(
        self,
        filtered_columns: List[Tuple[str, Operator]],
        order: List[Tuple[str, RequestedOrder]],
    ) -> int:
        cost = INITIAL_COST

        # if the operator is ``Operator.EQ`` we only need to fetch 1 day of data;
        # otherwise we potentially need to fetch "window" days of data
        for _, operator in filtered_columns:
            weight = 1 if operator == Operator.EQ else self.window
            cost += FETCHING_COST * weight

        return cost

    def get_data(  # pylint: disable=too-many-locals
        self,
        bounds: Dict[str, Filter],
        order: List[Tuple[str, RequestedOrder]],
    ) -> Iterator[Row]:
        # combine filters from the two time columns
        try:
            time_range = combine_time_filters(bounds)
        except ImpossibleFilterError:
            return

        today = date.today()
        first = today - timedelta(days=self.window - 1)
        start = time_range.start.date() if time_range.start else first
        end = time_range.end.date() if time_range.end else today
        _logger.debug("Range is %s to %s", start, end)

        # download data from every today from [start, end]
        while start <= end:
            url = "https://api.weatherapi.com/v1/history.json"
            params = {"key": self.api_key, "q": self.location, "dt": start}

            query_string = urllib.parse.urlencode(params)
            _logger.info("GET %s?%s", url, query_string)

            response = self._session.get(url, params=params)
            if response.ok:
                payload = response.json()
                local_timezone = dateutil.tz.gettz(payload["location"]["tz_id"])
                for record in payload["forecast"]["forecastday"][0]["hour"]:
                    row = {column: record[column] for column in self.get_columns()}
                    row["time"] = dateutil.parser.parse(record["time"]).replace(
                        tzinfo=local_timezone,
                    )
                    row["rowid"] = int(row["time_epoch"])
                    yield row
                    _logger.debug(row)

            start += timedelta(days=1)



"""
Fields representing columns of different types and capabilities.
"""
import datetime
from distutils.util import strtobool
from enum import Enum
from typing import Any, Generic, List, Optional, Type, TypeVar, cast

import dateutil.parser

from shillelagh.filters import Filter

Internal = TypeVar(
    "Internal",
    float,
    int,
    str,
    bool,
    datetime.date,
    datetime.datetime,
    datetime.time,
    datetime.timedelta,
    bytes,
)

External = TypeVar(
    "External",
    float,
    int,
    str,
    bool,
    datetime.date,
    datetime.datetime,
    datetime.time,
    datetime.timedelta,
    bytes,
)


class Order(Enum):
    """An enum for different orders a field can have."""

    # Use ASCENDING/DESCENDING when you have static data with 1+
    # columns pre-sorted. All other columns should have Order.NONE.
    ASCENDING = "ascending"
    DESCENDING = "descending"

    # Use NONE when you can't or don't want to sort the data. Sqlite
    # will then sort the provided data according to the query.
    NONE = "none"

    # Use ANY when the column can be sorted in any order. Usually
    # all other columns will also have Order.ANY.
    ANY = "any"


class Field(Generic[Internal, External]):

    """
    Represents a column in a table.

    A field is probably the most important concept in Shillelagh. Fields are
    used to annotate types, indicate which columns are filterable/sortable,
    and convert data between different format.

    1. Type annotation

    The most basic use of field is to indicate the types of columns in a
    given resource. For example, an adapter that connects to a database with
    two columns, a string and an integer, could look like this::

        class SimpleAdapter(Adapter):

            string_col = String()
            integer_col = Integer()

    For dynamic resources the columns might be generated dynamically, but
    the idea is the same::

        class DynamicAdapter(Adapter):

            type_map = {
                "int": Integer(),
                "float": Float(),
                "string": String(),
            }

            def get_columns(self) -> Dict[str, Field]:
                columns = read_columns()
                return {
                    column.name: type_map[column.type]
                    for column in columns
                }

    2. Filterable/sortable columns

    Most adapters can perform some kind of filtering/sorting on the data,
    return less data to the backend in order to optimize queries. Adapters
    indicate this on the fields, eg::

        class FilteringAdapter(Adapter):

            timestamp_col = DateTime(
                filters=[Equal, Range],
                order=Order.ANY,
                exact=True,
            )
            values_col = Float()

    The adapter above declares a column called ``timestamp_col`` that can
    be filtered using either an equality (`== '2020-01-01T00:00:00'`) or a
    range (`>= '2020-01-01T00:00:00'`). Because of this declaration the
    backend will delegate the filtering to the adapter, which will be
    responsible for translating the filters into API/filesystem calls to
    fulfill them.

    Additionally, the timestamp column also declares an order of ``ANY``, which
    means that the adapter can sort the data in any order requested by the
    backend. Fields can declare a static order (eg, ``Order.ASCENDING``) or
    no order at all.

    Finally, the field also indicates that the filtering is exact, and no
    post-filtering is needed to be done by the backend. In some cases it's
    useful to have adapters perform an initial coarse filtering (say, at
    the daily level), and have the backend perform the final fine filtering
    (say, at the second level), to simplify the logic in the adapter.

    3. Data conversion

    Fields are also responsible for converting data between different
    formats, so it can flow through layers. For example, an adapter might
    return booleans as strings, and these need to be converted back and
    forth to Python booleans. The ``StringBoolean`` field should be used in
    that case, and it will automatically convert the data to the format
    the adapter understands.

    Similarly, the APSW backend only accepts types understood by SQLite:
    ints, floats, strings, and bytes. This means that the backend needs to
    convert between, eg, native Python booleans and integers. This is also
    done by using the ``parse`` and ``format`` methods from fields (``IntBoolean``
    in this case).

    When creating new fields, the base class should declare the type of
    the "internal" representation (used by the adapter) and the "external"
    representation (native Python types). For example, if we have an
    adapter that stores numbers as strings we could define a new type::

        class StringNumber(Field[str, float]):  # Field[internal, external]
            type = "REAL"
            db_api_type = "NUMBER"

            # internal -> external
            def parse(self, value: Optional[str]) -> Optional[float]:
                return value if value is None else float(value)

            # external -> internal
            def format(self, value: Optional[float]) -> Optional[str]:
                return value if value is None else str(value)

    Then the adapter can declare columns using that field::

        class SomeAdapter(Adapter):

            number_col = StringNumber()

    With this, it can simply return rows with the number as a string,
    without having to explicitly do the conversion::

        {"rowid": 0, "number_col": "1.0"}

    """

    type = ""
    db_api_type = "DBAPIType"

    def __init__(
        self,
        filters: Optional[List[Type[Filter]]] = None,
        order: Order = Order.NONE,
        exact: bool = False,
    ):
        # a list of what kind of filters can be used on the column
        self.filters = filters or []

        # the ordering of the column
        self.order = order

        # are the results returned for the column exact or do they require
        # additional post-processing?
        self.exact = exact

    def __eq__(self, other: Any) -> bool:
        if other.__class__ != self.__class__:
            return NotImplemented

        return bool(
            self.filters == other.filters
            and self.order == other.order
            and self.exact == other.exact,
        )

    def parse(  # pylint: disable=no-self-use
        self,
        value: Optional[Internal],
    ) -> Optional[External]:
        """
        Convert from a DB type to a native Python type.

        Some databases might represent booleans as integers, or timestamps
        as strings. To convert those values to native Python types we call
        the ``parse`` method in the field associated with the column. Custom
        adapters can define their own derived fields to handle special
        formats.

        Eg, the Google Sheets API returns dates as strings in its response,
        using the format "Date(2018,0,1)" for "2018-01-01". A custom field
        allows the adapter to simply return the original value, and have it
        being automatically converted to a ``datetime.date`` object.

        This is not a staticmethod because some types need extra information
        in order to parse a value. Eg, GSheets takes into consideration the
        timezone of the sheet when parsing timestamps.
        """
        return cast(Optional[External], value)

    def format(  # pylint: disable=no-self-use
        self,
        value: Optional[External],
    ) -> Optional[Internal]:
        """
        Convert from a native Python type to a DB type.

        This should be the opposite of ``parse``.
        """
        return cast(Optional[Internal], value)

    def quote(self, value: Optional[Internal]) -> str:  # pylint: disable=no-self-use
        """
        Quote values.

        This method is used by some adapters to build a SQL expression.
        Eg, GSheets represents dates (and other time related types) with
        the prefix "date"::

            SELECT A, B WHERE C = date '2018-01-01'

        In order to handle that, the adapter defines its own time fields
        with custom ``quote`` methods.
        """
        if value is None:
            return "NULL"
        return str(value)


class Integer(Field[int, int]):
    """An integer."""

    type = "INTEGER"
    db_api_type = "NUMBER"


class RowID(Integer):
    """
    Custom field for the row ID.

    All Shillelagh adapters return a special column for the row ID. In
    many cases it's just an increasing integer, but it's used for DML
    in adapters that support it.
    """

    db_api_type = "ROWID"


class Float(Field[float, float]):
    """A float."""

    type = "REAL"
    db_api_type = "NUMBER"


class String(Field[str, str]):
    """A string."""

    type = "TEXT"
    db_api_type = "STRING"

    def quote(self, value: Optional[str]) -> str:
        if value is None:
            return "NULL"
        escaped_value = value.replace("'", "''")
        return f"'{escaped_value}'"


class Date(Field[datetime.date, datetime.date]):
    """
    A date.

    This field is used in adapters that use ``datetime.date`` as the
    internal representation of dates.
    """

    type = "DATE"
    db_api_type = "DATETIME"

    def quote(self, value: Optional[datetime.date]) -> str:
        if value is None:
            return "NULL"
        return f"'{value.isoformat()}'"


class ISODate(Field[str, datetime.date]):
    """
    A date.

    This field is used in adapters that use an ISO string as the
    internal representation of dates. SQLite, for example, has no
    concept of ``datetime.date`` objects, so we need to convert between
    the object and an ISO string when the data flows through SQLite.
    """

    type = "DATE"
    db_api_type = "DATETIME"

    def parse(self, value: Optional[str]) -> Optional[datetime.date]:
        if value is None:
            return None

        try:
            date = dateutil.parser.parse(value)
        except dateutil.parser.ParserError:
            return None

        return date.date()

    def format(self, value: Optional[datetime.date]) -> Optional[str]:
        if value is None:
            return None
        return value.isoformat()

    def quote(self, value: Optional[str]) -> str:
        if value is None:
            return "NULL"
        return f"'{value}'"


class Time(Field[datetime.time, datetime.time]):
    """
    A time of the day.

    This field is used in adapters that use ``datetime.time`` as the
    internal representation of times of the day.
    """

    type = "TIME"
    db_api_type = "DATETIME"

    def quote(self, value: Optional[datetime.time]) -> str:
        if value is None:
            return "NULL"
        return f"'{value.isoformat()}'"


class ISOTime(Field[str, datetime.time]):
    """
    A time of the day.

    This field is used in adapters that use an ISO string as the
    internal representation of dates. SQLite, for example, has no
    concept of ``datetime.time`` objects, so we need to convert between
    the object and an ISO string when the data flows through SQLite.
    """

    type = "TIME"
    db_api_type = "DATETIME"

    def parse(self, value: Optional[str]) -> Optional[datetime.time]:
        if value is None:
            return None

        try:
            timestamp = dateutil.parser.parse(value)
        except dateutil.parser.ParserError:
            return None

        time = timestamp.time()

        # timezone is not preserved
        return time.replace(tzinfo=timestamp.tzinfo)

    def format(self, value: Optional[datetime.time]) -> Optional[str]:
        if value is None:
            return None
        return value.isoformat()

    def quote(self, value: Optional[str]) -> str:
        if value is None:
            return "NULL"
        return f"'{value}'"


class DateTime(Field[datetime.datetime, datetime.datetime]):
    """
    A timestamp.

    This field is used in adapters that use ``datetime.datetime``
    as the internal representation of timestamps.
    """

    type = "TIMESTAMP"
    db_api_type = "DATETIME"

    def quote(self, value: Optional[datetime.datetime]) -> str:
        if value is None:
            return "NULL"
        return f"'{value.isoformat()}'"


class ISODateTime(Field[str, datetime.datetime]):
    """
    A timestamp.

    This field is used in adapters that use an ISO string as the
    internal representation of dates. SQLite, for example, has no
    concept of ``datetime.datetime`` objects, so we need to convert
    between the object and an ISO string when the data flows
    through SQLite.
    """

    type = "TIMESTAMP"
    db_api_type = "DATETIME"

    def parse(self, value: Optional[str]) -> Optional[datetime.datetime]:
        if value is None:
            return None

        try:
            timestamp = dateutil.parser.parse(value)
        except dateutil.parser.ParserError:
            return None

        # if the timestamp has a timezone change it to UTC, so that
        # timestamps in different timezones can be compared as strings
        if timestamp.tzinfo is not None:
            timestamp = timestamp.astimezone(datetime.timezone.utc)

        return timestamp

    def format(self, value: Optional[datetime.datetime]) -> Optional[str]:
        if value is None:
            return None

        # if the timestamp has a timezone change it to UTC, so that
        # timestamps in different timezones can be compared as strings
        if value.tzinfo is not None:
            value = value.astimezone(datetime.timezone.utc)

        return value.isoformat()

    def quote(self, value: Optional[str]) -> str:
        if value is None:
            return "NULL"
        return f"'{value}'"


class StringDuration(Field[str, datetime.timedelta]):
    """
    A duration.

    This field represents durations as a string.
    """

    type = "DURATION"
    db_api_type = "DATETIME"

    def parse(self, value: Optional[str]) -> Optional[datetime.timedelta]:
        if value is None:
            return None

        if ", " in value:
            rest, value = value.split(", ")
            days = int(rest.split(" ")[0])
        else:
            days = 0

        pattern = "%H:%M:%S.%f" if "." in value else "%H:%M:%S"

        return (
            datetime.datetime.strptime(value, pattern)
            - datetime.datetime(
                1900,
                1,
                1,
            )
            + datetime.timedelta(days=days)
        )

    def format(self, value: Optional[datetime.timedelta]) -> Optional[str]:
        if value is None:
            return None

        return str(value)

    def quote(self, value: Optional[str]) -> str:
        if value is None:
            return "NULL"
        return f"'{value}'"


class Blob(Field[bytes, bytes]):
    """
    A blob of bytes.

    This is used to represent binary data.
    """

    type = "BLOB"
    db_api_type = "BINARY"

    def quote(self, value: Optional[bytes]) -> str:
        if value is None:
            return "NULL"
        return f"X'{value.hex()}'"


class StringBlob(Field[str, bytes]):
    """
    A blob of bytes.

    This field is used in adapters that represent binary data as a
    string with the hexadecimal representation of the bytes.
    """

    type = "BLOB"
    db_api_type = "BINARY"

    def parse(self, value: Optional[str]) -> Optional[bytes]:
        if value is None:
            return None
        return bytes.fromhex(value)

    def format(self, value: Optional[bytes]) -> Optional[str]:
        if value is None:
            return None
        return cast(str, value.hex())

    def quote(self, value: Optional[str]) -> str:
        if value is None:
            return "NULL"
        return f"X'{value}'"


class Boolean(Field[bool, bool]):
    """A boolean."""

    type = "BOOLEAN"
    db_api_type = "NUMBER"

    def quote(self, value: Optional[bool]) -> str:
        if value is None:
            return "NULL"
        return "TRUE" if value else "FALSE"


class StringBoolean(Field[str, bool]):
    """
    A boolean.

    This field is used in adapters that represent booleans as strings,
    eg, "TRUE" and "FALSE".
    """

    type = "BOOLEAN"
    db_api_type = "NUMBER"

    def parse(self, value: Optional[str]) -> Optional[bool]:
        if value is None:
            return None
        return bool(strtobool(value))

    def format(self, value: Optional[bool]) -> Optional[str]:
        if value is None:
            return None
        return "TRUE" if value else "FALSE"

    def quote(self, value: Optional[str]) -> str:
        if value is None:
            return "NULL"
        return value


class IntBoolean(Field[int, bool]):
    """
    A boolean.

    This field is used in adapters that represent booleans as an
    integer. SQLite, eg, has no boolean type, using 1 and 0 to
    represent true and false, respectively.
    """

    type = "BOOLEAN"
    db_api_type = "NUMBER"

    def parse(self, value: Optional[int]) -> Optional[bool]:
        if value is None:
            return None
        return bool(value)

    def format(self, value: Optional[bool]) -> Optional[int]:
        if value is None:
            return None
        return 1 if value else 0

    def quote(self, value: Optional[int]) -> str:
        if value is None:
            return "NULL"
        return str(value)



"""
An adapter for GitHub.
"""
import logging
import urllib.parse
from dataclasses import dataclass
from typing import Any, Dict, Iterator, List, Optional, Tuple

import requests_cache
from jsonpath import JSONPath

from shillelagh.adapters.base import Adapter
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import Boolean, Field, Integer, ISODateTime, String
from shillelagh.filters import Equal, Filter
from shillelagh.typing import RequestedOrder, Row

_logger = logging.getLogger(__name__)

PAGE_SIZE = 100


@dataclass
class Column:
    """
    A class to track columns in GitHub resources.
    """

    # The name of the column in the virtual table, eg, ``userid``.
    name: str

    # The JSON path to the value in the GitHub response. Eg, user ID is
    # returned as ``{"user": {"id": 42}}``, so the JSON path would be
    # ``user.id``.
    json_path: str

    # The Shillelagh field to be used for the column.
    field: Field

    # A default value for when the column is not specified. Eg, for ``pulls``
    # the API defaults to show only PRs with an open state, so we need to
    # default the column to ``all`` to fetch all PRs when state is not
    # specified in the query.
    default: Optional[Filter] = None


# a mapping from the column name (eg, ``userid``) to the path in the JSON
# response (``{"user": {"id": 42}}`` => ``user.id``) together with the field
TABLES: Dict[str, Dict[str, List[Column]]] = {
    "repos": {
        "pulls": [
            Column("url", "html_url", String()),
            Column("id", "id", Integer()),
            Column("number", "number", Integer(filters=[Equal])),
            Column("state", "state", String(filters=[Equal]), Equal("all")),
            Column("title", "title", String()),
            Column("userid", "user.id", Integer()),
            Column("username", "user.login", Integer()),
            Column("draft", "draft", Boolean()),
            Column("head", "head.ref", String(filters=[Equal])),  # head.label?
            Column("created_at", "created_at", ISODateTime()),
            Column("updated_at", "updated_at", ISODateTime()),
            Column("closed_at", "closed_at", ISODateTime()),
            Column("merged_at", "merged_at", ISODateTime()),
        ],
    },
}


class GitHubAPI(Adapter):

    """
    An adapter for GitHub.
    """

    safe = True

    @staticmethod
    def supports(uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        parsed = urllib.parse.urlparse(uri)

        if parsed.path.count("/") != 4:
            return False

        # pylint: disable=unused-variable
        _, base, owner, repo, resource = parsed.path.rsplit("/", 4)
        return (
            parsed.netloc == "api.github.com"
            and base in TABLES
            and resource in TABLES[base]
        )

    @staticmethod
    def parse_uri(uri: str) -> Tuple[str, str, str, str]:
        parsed = urllib.parse.urlparse(uri)
        _, base, owner, repo, resource = parsed.path.rsplit("/", 4)
        return (
            base,
            owner,
            repo,
            resource,
        )

    def __init__(  # pylint: disable=too-many-arguments
        self,
        base: str,
        owner: str,
        repo: str,
        resource: str,
        access_token: Optional[str] = None,
    ):
        super().__init__()

        self.base = base
        self.owner = owner
        self.repo = repo
        self.resource = resource
        self.access_token = access_token

        # use a cache for the API requests
        self._session = requests_cache.CachedSession(
            cache_name="github_cache",
            backend="sqlite",
            expire_after=180,
        )

    def get_columns(self) -> Dict[str, Field]:
        return {
            column.name: column.field for column in TABLES[self.base][self.resource]
        }

    def get_data(
        self,
        bounds: Dict[str, Filter],
        order: List[Tuple[str, RequestedOrder]],
    ) -> Iterator[Row]:
        # apply default values
        for column in TABLES[self.base][self.resource]:
            if column.default is not None and column.name not in bounds:
                bounds[column.name] = column.default

        if "number" in bounds:
            number = bounds.pop("number").value  # type: ignore
            return self._get_single_resource(number)

        return self._get_multiple_resources(bounds)

    def _get_single_resource(self, number: int) -> Iterator[Row]:
        """
        Return a specific resource.
        """
        headers = {"Accept": "application/vnd.github.v3+json"}
        if self.access_token:
            headers["Authorization"] = f"Bearer {self.access_token}"

        url = (
            f"https://api.github.com/{self.base}/{self.owner}/"
            f"{self.repo}/{self.resource}/{number}"
        )

        _logger.info("GET %s", url)
        response = self._session.get(url, headers=headers)
        payload = response.json()

        row = {
            column.name: JSONPath(column.json_path).parse(payload)[0]
            for column in TABLES[self.base][self.resource]
        }
        row["rowid"] = 0
        _logger.debug(row)
        yield row

    def _get_multiple_resources(self, bounds: Dict[str, Filter]) -> Iterator[Row]:
        """
        Return multiple resources.
        """
        headers = {"Accept": "application/vnd.github.v3+json"}
        if self.access_token:
            headers["Authorization"] = f"Bearer {self.access_token}"

        url = f"https://api.github.com/{self.base}/{self.owner}/{self.repo}/{self.resource}"

        # map filters in ``bounds`` to query params
        params = {name: filter_.value for name, filter_ in bounds.items()}  # type: ignore
        params["per_page"] = PAGE_SIZE

        page = 1
        while True:
            _logger.info("GET %s (page %d)", url, page)
            params["page"] = page
            response = self._session.get(url, headers=headers, params=params)

            payload = response.json()
            if not payload:
                break

            if not response.ok:
                raise ProgrammingError(payload["message"])

            for i, resource in enumerate(payload):
                row = {
                    column.name: JSONPath(column.json_path).parse(resource)[0]
                    for column in TABLES[self.base][self.resource]
                }
                row["rowid"] = i + (page - 1) * PAGE_SIZE
                _logger.debug(row)
                yield row

            page += 1



"""
Filters for representing SQL predicates.
"""
import re
from enum import Enum
from typing import Any, Optional, Set, Tuple


class Operator(Enum):
    """
    Enum representing support comparisons.
    """

    EQ = "=="
    NE = "!="
    GE = ">="
    GT = ">"
    LE = "<="
    LT = "<"
    IS_NULL = "IS NULL"
    IS_NOT_NULL = "IS NOT NULL"
    LIKE = "LIKE"


class Side(Enum):
    """Define the side of an interval endpoint."""

    LEFT = "LEFT"
    RIGHT = "RIGHT"


class Endpoint:
    """
    One of the two endpoints of a ``Range``.

    Used to compare ranges. Eg, the range ``>10`` can be represented by:

        >>> start = Endpoint(10, False, Side.LEFT)
        >>> end = Endpoint(None, True, Side.RIGHT)
        >>> print(f'{start},{end}')
        (10,∞]

    The first endpoint represents the value 10 at the left side, in an open
    interval. The second endpoint represents infinity in this case.
    """

    def __init__(self, value: Any, include: bool, side: Side):
        self.value = value
        self.include = include
        self.side = side

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Endpoint):
            return NotImplemented

        return self.value == other.value and self.include == other.include

    def __gt__(self, other: Any) -> bool:  # pylint: disable=too-many-return-statements
        if not isinstance(other, Endpoint):
            return NotImplemented

        if self.value is None:
            return self.side == Side.RIGHT

        if other.value is None:
            return other.side == Side.LEFT

        if self.value == other.value:
            if self.side == Side.LEFT:
                if other.side == Side.LEFT:
                    return not self.include and other.include
                return not self.include

            # self.side = Side.RIGHT
            if other.side == Side.RIGHT:
                return not other.include and self.include
            return False

        return bool(self.value > other.value)

    # needed for ``max()``
    def __lt__(self, other: Any) -> bool:
        return not self > other

    def __str__(self) -> str:
        """
        Representation of an endpoint.

            >>> print(Endpoint(10, False, Side.LEFT))
            (10

        """
        if self.side == Side.LEFT:
            symbol = "[" if self.include else "("
            value = "-∞" if self.value is None else self.value
            return f"{symbol}{value}"

        symbol = "]" if self.include else ")"
        value = "∞" if self.value is None else self.value
        return f"{value}{symbol}"


def get_endpoints_from_operation(
    operator: Operator,
    value: Any,
) -> Tuple[Endpoint, Endpoint]:
    """
    Returns endpoints from an operation.
    """
    if operator == Operator.EQ:
        return Endpoint(value, True, Side.LEFT), Endpoint(value, True, Side.RIGHT)
    if operator == Operator.GE:
        return Endpoint(value, True, Side.LEFT), Endpoint(None, True, Side.RIGHT)
    if operator == Operator.GT:
        return Endpoint(value, False, Side.LEFT), Endpoint(None, True, Side.RIGHT)
    if operator == Operator.LE:
        return Endpoint(None, True, Side.LEFT), Endpoint(value, True, Side.RIGHT)
    if operator == Operator.LT:
        return Endpoint(None, True, Side.LEFT), Endpoint(value, False, Side.RIGHT)

    raise Exception(f"Invalid operator: {operator}")


class Filter:
    """
    A filter representing a SQL predicate.
    """

    operators: Set[Operator] = set()

    @classmethod
    def build(cls, operations: Set[Tuple[Operator, Any]]) -> "Filter":
        """
        Given a set of operations, build a filter:

            >>> operations = [(Operator.GT, 10), (Operator.GT, 20)]
            >>> print(Range.build(operations))
            >20

        """
        raise NotImplementedError("Subclass must implement ``build``")

    def check(self, value: Any) -> bool:
        """
        Test if a given filter matches a value:

            >>> operations = [(Operator.GT, 10), (Operator.GT, 20)]
            >>> filter_ = Range.build(operations)
            >>> filter_.check(10)
            False
            >>> filter_.check(30)
            True

        """
        raise NotImplementedError("Subclass must implement ``check``")


class Impossible(Filter):
    """
    Custom Filter returned when impossible conditions are passed.
    """

    @classmethod
    def build(cls, operations: Set[Tuple[Operator, Any]]) -> Filter:
        return Impossible()

    def check(self, value: Any) -> bool:
        return False

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Impossible):
            return NotImplemented

        return True


class IsNull(Filter):
    """
    Filter for ``IS NULL``.
    """

    operators: Set[Operator] = {Operator.IS_NULL}

    @classmethod
    def build(cls, operations: Set[Tuple[Operator, Any]]) -> Filter:
        return IsNull()

    def check(self, value: Any) -> bool:
        return value is None

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, IsNull):
            return NotImplemented

        return True

    def __repr__(self) -> str:
        return "IS NULL"


class IsNotNull(Filter):
    """
    Filter for ``IS NOT NULL``.
    """

    operators: Set[Operator] = {Operator.IS_NOT_NULL}

    @classmethod
    def build(cls, operations: Set[Tuple[Operator, Any]]) -> Filter:
        return IsNotNull()

    def check(self, value: Any) -> bool:
        return value is not None

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, IsNotNull):
            return NotImplemented

        return True

    def __repr__(self) -> str:
        return "IS NOT NULL"


class Equal(Filter):
    """
    Equality comparison.
    """

    operators: Set[Operator] = {
        Operator.EQ,
    }

    def __init__(self, value: Any):
        self.value = value

    @classmethod
    def build(cls, operations: Set[Tuple[Operator, Any]]) -> Filter:
        values = {value for operator, value in operations}
        if len(values) != 1:
            return Impossible()

        return cls(values.pop())

    def check(self, value: Any) -> bool:
        return bool(value == self.value)

    def __repr__(self) -> str:
        return f"=={self.value}"


class NotEqual(Filter):
    """
    Inequality comparison.
    """

    operators: Set[Operator] = {
        Operator.NE,
    }

    def __init__(self, value: Any):
        self.value = value

    @classmethod
    def build(cls, operations: Set[Tuple[Operator, Any]]) -> Filter:
        values = {value for operator, value in operations}
        if len(values) != 1:
            return Impossible()

        return cls(values.pop())

    def check(self, value: Any) -> bool:
        return bool(value != self.value)

    def __repr__(self) -> str:
        return f"!={self.value}"


class Like(Filter):
    """
    Substring searches.
    """

    operators: Set[Operator] = {
        Operator.LIKE,
    }

    def __init__(self, value: Any):
        self.value = value
        self.regex = re.compile(
            self.value.replace("_", ".").replace("%", ".*"),
            re.IGNORECASE,
        )

    @classmethod
    def build(cls, operations: Set[Tuple[Operator, Any]]) -> Filter:
        # we only accept a single value
        values = {value for operator, value in operations}
        if len(values) != 1:
            return Impossible()

        return cls(values.pop())

    def check(self, value: Any) -> bool:
        return bool(self.regex.match(value))

    def __repr__(self) -> str:
        return f"LIKE {self.value}"


class Range(Filter):
    """
    A range comparison.

    This filter represents a range, with an optional start and an
    optional end. Start and end can be inclusive or exclusive.

    Ranges can be combined by adding them:

        >>> range1 = Range(start=10)
        >>> range2 = Range(start=20)
        >>> print(range1 + range2)
        >20
        >>> range3 = Range(end=40)
        >>> print(range2 + range3)
        >20,<40

    """

    def __init__(
        self,
        start: Optional[Any] = None,
        end: Optional[Any] = None,
        include_start: bool = False,
        include_end: bool = False,
    ):
        self.start = start
        self.end = end
        self.include_start = include_start
        self.include_end = include_end

    operators: Set[Operator] = {
        Operator.EQ,
        Operator.GE,
        Operator.GT,
        Operator.LE,
        Operator.LT,
    }

    def __eq__(self, other: Any):
        if not isinstance(other, Range):
            return NotImplemented

        return (
            self.start == other.start
            and self.end == other.end
            and self.include_start == other.include_start
            and self.include_end == other.include_end
        )

    def __add__(self, other: Any) -> Filter:
        if not isinstance(other, Range):
            return NotImplemented

        start = Endpoint(self.start, self.include_start, Side.LEFT)
        end = Endpoint(self.end, self.include_end, Side.RIGHT)

        new_start = Endpoint(other.start, other.include_start, Side.LEFT)
        new_end = Endpoint(other.end, other.include_end, Side.RIGHT)

        start = max(start, new_start)
        end = min(end, new_end)

        if start > end:
            return Impossible()

        return Range(start.value, end.value, start.include, end.include)

    @classmethod
    def build(cls, operations: Set[Tuple[Operator, Any]]) -> Filter:
        start = Endpoint(None, True, Side.LEFT)
        end = Endpoint(None, True, Side.RIGHT)

        for operator, value in operations:
            new_start, new_end = get_endpoints_from_operation(operator, value)

            start = max(start, new_start)
            end = min(end, new_end)

            if start > end:
                return Impossible()

        return cls(start.value, end.value, start.include, end.include)

    def check(self, value: Any) -> bool:
        if self.start is not None:
            if self.include_start and value < self.start:
                return False
            if not self.include_start and value <= self.start:
                return False

        if self.end is not None:
            if self.include_end and value > self.end:
                return False
            if not self.include_end and value >= self.end:
                return False

        return True

    def __str__(self) -> str:
        if self.start == self.end and self.include_start and self.include_end:
            return f"=={self.start}"

        comparisons = []
        if self.start is not None:
            operator = ">=" if self.include_start else ">"
            comparisons.append(f"{operator}{self.start}")
        if self.end is not None:
            operator = "<=" if self.include_end else "<"
            comparisons.append(f"{operator}{self.end}")
        return ",".join(comparisons)



"""
Tests for the Socrata adapter.
"""
from datetime import date

import pytest
from pytest_mock import MockerFixture
from requests import Session
from requests_mock.mocker import Mocker

from shillelagh.adapters.api.socrata import Number, SocrataAPI
from shillelagh.backends.apsw.db import connect
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import Order
from shillelagh.filters import Impossible, Operator

from ...fakes import cdc_data_response, cdc_metadata_response


def test_socrata(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test a simple query.
    """
    mocker.patch(
        "shillelagh.adapters.api.socrata.requests_cache.CachedSession",
        return_value=Session(),
    )

    metadata_url = "https://data.cdc.gov/api/views/unsk-b7fc"
    requests_mock.get(metadata_url, json=cdc_metadata_response)

    data_url = (
        "https://data.cdc.gov/resource/unsk-b7fc.json?"
        "%24query=SELECT+%2A+WHERE+location+%3D+%27US%27+ORDER+BY+date+DESC"
    )
    requests_mock.get(data_url, json=cdc_data_response)

    connection = connect(":memory:")
    cursor = connection.cursor()
    sql = """
        SELECT date, administered_dose1_recip_4
        FROM "https://data.cdc.gov/resource/unsk-b7fc.json"
        WHERE location = 'US'
        ORDER BY date DESC
        LIMIT 7
    """
    data = list(cursor.execute(sql))
    assert data == [
        (date(2021, 6, 3), 63.0),
        (date(2021, 6, 2), 62.9),
        (date(2021, 6, 1), 62.8),
        (date(2021, 5, 31), 62.7),
        (date(2021, 5, 30), 62.6),
        (date(2021, 5, 29), 62.4),
        (date(2021, 5, 28), 62.2),
    ]


def test_socrata_app_token_url(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test app token being passed via the URL.
    """
    mocker.patch(
        "shillelagh.adapters.api.socrata.requests_cache.CachedSession",
        return_value=Session(),
    )

    metadata_url = "https://data.cdc.gov/api/views/unsk-b7fc"
    requests_mock.get(metadata_url, json=cdc_metadata_response)

    data_url = (
        "https://data.cdc.gov/resource/unsk-b7fc.json?"
        "%24query=SELECT+%2A+WHERE+location+%3D+%27OK%27+ORDER+BY+date+DESC"
    )
    data = requests_mock.get(data_url, json=cdc_data_response)

    connection = connect(":memory:")
    cursor = connection.cursor()
    sql = """
        SELECT date, administered_dose1_recip_4
        FROM "https://data.cdc.gov/resource/unsk-b7fc.json?$$app_token=XXX"
        WHERE location = 'OK'
        ORDER BY date DESC
        LIMIT 7
    """
    cursor.execute(sql)
    assert data.last_request.headers == {"X-App-Token": "XXX"}


def test_socrata_app_token_connection(
    mocker: MockerFixture,
    requests_mock: Mocker,
) -> None:
    """
    Test app token being passed via the connection instead of the URL.
    """
    mocker.patch(
        "shillelagh.adapters.api.socrata.requests_cache.CachedSession",
        return_value=Session(),
    )

    metadata_url = "https://data.cdc.gov/api/views/unsk-b7fc"
    requests_mock.get(metadata_url, json=cdc_metadata_response)

    data_url = (
        "https://data.cdc.gov/resource/unsk-b7fc.json?"
        "%24query=SELECT+%2A+WHERE+location+%3D+%27NY%27+ORDER+BY+date+DESC"
    )
    data = requests_mock.get(data_url, json=cdc_data_response)
    connection = connect(
        ":memory:",
        adapter_kwargs={"socrataapi": {"app_token": "YYY"}},
    )
    cursor = connection.cursor()
    sql = """
        SELECT date, administered_dose1_recip_4
        FROM "https://data.cdc.gov/resource/unsk-b7fc.json"
        WHERE location = 'NY'
        ORDER BY date DESC
        LIMIT 7
    """
    cursor.execute(sql)
    assert data.last_request.headers == {"X-App-Token": "YYY"}


def test_socrata_no_data(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test that some queries return no data.
    """
    mocker.patch(
        "shillelagh.adapters.api.socrata.requests_cache.CachedSession",
        return_value=Session(),
    )

    metadata_url = "https://data.cdc.gov/api/views/unsk-b7fc"
    requests_mock.get(metadata_url, json=cdc_metadata_response)

    data_url = (
        "https://data.cdc.gov/resource/unsk-b7fc.json?"
        "%24query=SELECT+%2A+WHERE+location+%3D+%27BR%27+ORDER+BY+date+DESC"
    )
    requests_mock.get(data_url, json=[])

    connection = connect(":memory:")
    cursor = connection.cursor()
    sql = """
        SELECT date, administered_dose1_recip_4
        FROM "https://data.cdc.gov/resource/unsk-b7fc.json"
        WHERE location = 'BR'
        ORDER BY date DESC
        LIMIT 7
    """
    data = list(cursor.execute(sql))
    assert data == []


def test_socrata_impossible(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test that impossible queries return no data.
    """
    mocker.patch(
        "shillelagh.adapters.api.socrata.requests_cache.CachedSession",
        return_value=Session(),
    )

    metadata_url = "https://data.cdc.gov/api/views/unsk-b7fc"
    requests_mock.get(metadata_url, json=cdc_metadata_response)

    connection = connect(":memory:")
    cursor = connection.cursor()
    sql = """
        SELECT date, administered_dose1_recip_4
        FROM "https://data.cdc.gov/resource/unsk-b7fc.json"
        WHERE location = 'US' AND location = 'AZ'
        ORDER BY date DESC
        LIMIT 7
    """
    data = list(cursor.execute(sql))
    assert data == []

    # test for apsw 3.36, since it doesn't call ``get_data`` when the
    # query is impossible to resolve
    adapter = SocrataAPI("data.cdc.gov", "unsk-b7fc")
    assert list(adapter.get_data({"location": Impossible()}, [])) == []


def test_socrata_invalid_query(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test that invalid queries are handled correctly.
    """
    mocker.patch(
        "shillelagh.adapters.api.socrata.requests_cache.CachedSession",
        return_value=Session(),
    )

    metadata_url = "https://data.cdc.gov/api/views/unsk-b7fc"
    requests_mock.get(metadata_url, json=cdc_metadata_response)

    data_url = (
        "https://data.cdc.gov/resource/unsk-b7fc.json?"
        "%24query=SELECT+%2A+WHERE+location+%3D+%27CA%27+ORDER+BY+date+DESC"
    )
    requests_mock.get(
        data_url,
        json={
            "message": "Invalid SoQL query",
            "errorCode": "query.soql.invalid",
            "data": {},
        },
    )

    connection = connect(":memory:")
    cursor = connection.cursor()
    sql = """
        SELECT date, administered_dose1_recip_4
        FROM "https://data.cdc.gov/resource/unsk-b7fc.json"
        WHERE location = 'CA'
        ORDER BY date DESC
        LIMIT 7
    """
    with pytest.raises(ProgrammingError) as excinfo:
        cursor.execute(sql)
    assert str(excinfo.value) == "Invalid SoQL query"


def test_number() -> None:
    """
    Test that numbers are converted correctly.
    """
    assert Number().parse("1.0") == 1.0
    assert Number().parse(None) is None
    assert Number().format(1.0) == "1.0"
    assert Number().format(None) is None
    assert Number().quote("1.0") == "1.0"
    assert Number().quote(None) == "NULL"


@pytest.mark.integration_test
def test_integration(adapter_kwargs) -> None:
    """
    Test fetching data from the CDC.
    """
    connection = connect(":memory:", adapter_kwargs=adapter_kwargs)
    cursor = connection.cursor()

    sql = """
        SELECT administered_dose1_recip_4
        FROM "https://data.cdc.gov/resource/unsk-b7fc.json"
        WHERE location = ? AND date = ?
    """
    cursor.execute(sql, ("US", date(2021, 7, 4)))
    assert cursor.fetchall() == [(67.1,)]


def test_get_cost(mocker: MockerFixture) -> None:
    """
    Test ``get_cost``.
    """
    mocker.patch(
        "shillelagh.adapters.api.socrata.requests_cache.CachedSession",
    )

    adapter = SocrataAPI("netloc", "dataset", "XXX")

    assert adapter.get_cost([], []) == 0
    assert adapter.get_cost([("one", Operator.EQ)], []) == 1000
    assert adapter.get_cost([("one", Operator.EQ), ("two", Operator.GT)], []) == 2000
    assert (
        adapter.get_cost(
            [("one", Operator.EQ), ("two", Operator.GT)],
            [("one", Order.ASCENDING)],
        )
        == 11965
    )
    assert (
        adapter.get_cost(
            [("one", Operator.EQ), ("two", Operator.GT)],
            [("one", Order.ASCENDING), ("two", Order.DESCENDING)],
        )
        == 21931
    )



# pylint: disable=too-many-lines
"""
Tests for the Datasette adapter.
"""
import datetime

import pytest
from pytest_mock import MockerFixture
from requests import Session
from requests_mock.mocker import Mocker

from shillelagh.backends.apsw.db import connect
from shillelagh.exceptions import ProgrammingError

from ...fakes import github_response, github_single_response


def test_github(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test a simple request.
    """
    mocker.patch(
        "shillelagh.adapters.api.github.requests_cache.CachedSession",
        return_value=Session(),
    )

    page1_url = "https://api.github.com/repos/apache/superset/pulls?state=all&per_page=100&page=1"
    requests_mock.get(page1_url, json=github_response)
    page2_url = "https://api.github.com/repos/apache/superset/pulls?state=all&per_page=100&page=2"
    requests_mock.get(page2_url, json=[])

    connection = connect(":memory:")
    cursor = connection.cursor()

    sql = """
        SELECT * FROM
        "https://api.github.com/repos/apache/superset/pulls"
    """
    data = list(cursor.execute(sql))
    assert data == [
        (
            "https://github.com/apache/superset/pull/16581",
            726927278,
            16581,
            "open",
            "feat: Arash/datasets and reports",
            48933336,
            "AAfghahi",
            False,
            "arash/datasetsAndReports",
            datetime.datetime(2021, 9, 3, 15, 57, 37, tzinfo=datetime.timezone.utc),
            datetime.datetime(2021, 9, 3, 15, 57, 39, tzinfo=datetime.timezone.utc),
            None,
            None,
        ),
        (
            "https://github.com/apache/superset/pull/16576",
            726586317,
            16576,
            "open",
            "chore(viz): remove legacy table viz code",
            33317356,
            "villebro",
            True,
            "villebro/remove-table-viz",
            datetime.datetime(2021, 9, 3, 7, 52, 18, tzinfo=datetime.timezone.utc),
            datetime.datetime(2021, 9, 3, 8, 48, 45, tzinfo=datetime.timezone.utc),
            None,
            None,
        ),
        (
            "https://github.com/apache/superset/pull/16571",
            726278427,
            16571,
            "open",
            "chore(deps-dev): bump storybook-addon-jsx from 7.3.3 to 7.3.13 in /superset-frontend",
            49699333,
            "dependabot[bot]",
            False,
            "dependabot/npm_and_yarn/superset-frontend/storybook-addon-jsx-7.3.13",
            datetime.datetime(2021, 9, 2, 20, 51, 50, tzinfo=datetime.timezone.utc),
            datetime.datetime(2021, 9, 2, 21, 22, 54, tzinfo=datetime.timezone.utc),
            None,
            None,
        ),
        (
            "https://github.com/apache/superset/pull/16569",
            726107410,
            16569,
            "open",
            "docs: versioned _export Stable",
            47772523,
            "amitmiran137",
            False,
            "version_export_ff_on",
            datetime.datetime(2021, 9, 2, 16, 52, 34, tzinfo=datetime.timezone.utc),
            datetime.datetime(2021, 9, 2, 18, 6, 27, tzinfo=datetime.timezone.utc),
            None,
            None,
        ),
        (
            "https://github.com/apache/superset/pull/16566",
            725808286,
            16566,
            "open",
            "fix(docker): add ecpg to docker image",
            33317356,
            "villebro",
            False,
            "villebro/libecpg",
            datetime.datetime(2021, 9, 2, 12, 1, 2, tzinfo=datetime.timezone.utc),
            datetime.datetime(2021, 9, 2, 12, 6, 50, tzinfo=datetime.timezone.utc),
            None,
            None,
        ),
        (
            "https://github.com/apache/superset/pull/16564",
            725669631,
            16564,
            "open",
            "refactor: orderby control refactoring",
            2016594,
            "zhaoyongjie",
            True,
            "refactor_orderby",
            datetime.datetime(2021, 9, 2, 9, 45, 40, tzinfo=datetime.timezone.utc),
            datetime.datetime(2021, 9, 3, 10, 31, 4, tzinfo=datetime.timezone.utc),
            None,
            None,
        ),
        (
            "https://github.com/apache/superset/pull/16554",
            724863880,
            16554,
            "open",
            "refactor: Update async query init to support runtime feature flags",
            296227,
            "robdiciuccio",
            False,
            "rd/async-query-init-refactor",
            datetime.datetime(2021, 9, 1, 19, 51, 51, tzinfo=datetime.timezone.utc),
            datetime.datetime(2021, 9, 1, 22, 29, 46, tzinfo=datetime.timezone.utc),
            None,
            None,
        ),
        (
            "https://github.com/apache/superset/pull/16549",
            724669525,
            16549,
            "open",
            "feat(dashboard): Native filters - add type to native filter configuration",
            12539911,
            "m-ajay",
            False,
            "feat/migration-add-type-to-native-filter",
            datetime.datetime(2021, 9, 1, 16, 35, 50, tzinfo=datetime.timezone.utc),
            datetime.datetime(2021, 9, 3, 17, 33, 42, tzinfo=datetime.timezone.utc),
            None,
            None,
        ),
        (
            "https://github.com/apache/superset/pull/16548",
            724668038,
            16548,
            "open",
            "refactor: sql_json view endpoint: encapsulate ctas parameters",
            35701650,
            "ofekisr",
            False,
            "refactor/sql_json_view4",
            datetime.datetime(2021, 9, 1, 16, 33, 45, tzinfo=datetime.timezone.utc),
            datetime.datetime(2021, 9, 1, 17, 6, 32, tzinfo=datetime.timezone.utc),
            None,
            None,
        ),
        (
            "https://github.com/apache/superset/pull/16545",
            724509555,
            16545,
            "open",
            "perf(dashboard): decrease number of rerenders of FiltersBadge",
            15073128,
            "kgabryje",
            False,
            "perf/dashboard-rerenders-4",
            datetime.datetime(2021, 9, 1, 13, 41, 12, tzinfo=datetime.timezone.utc),
            datetime.datetime(2021, 9, 2, 15, 39, 15, tzinfo=datetime.timezone.utc),
            None,
            None,
        ),
    ]


def test_github_single_resource(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test a request to a single resource.
    """
    mocker.patch(
        "shillelagh.adapters.api.github.requests_cache.CachedSession",
        return_value=Session(),
    )

    url = "https://api.github.com/repos/apache/superset/pulls/16581"
    requests_mock.get(url, json=github_single_response)

    connection = connect(":memory:")
    cursor = connection.cursor()

    sql = """
        SELECT * FROM
        "https://api.github.com/repos/apache/superset/pulls"
        WHERE number = 16581
    """
    data = list(cursor.execute(sql))
    assert data == [
        (
            "https://github.com/apache/superset/pull/16581",
            726927278,
            16581,
            "open",
            "feat: Arash/datasets and reports",
            48933336,
            "AAfghahi",
            False,
            "arash/datasetsAndReports",
            datetime.datetime(2021, 9, 3, 15, 57, 37, tzinfo=datetime.timezone.utc),
            datetime.datetime(2021, 9, 3, 15, 57, 39, tzinfo=datetime.timezone.utc),
            None,
            None,
        ),
    ]


def test_github_rate_limit(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test that the adapter was rate limited by the API.
    """
    mocker.patch(
        "shillelagh.adapters.api.github.requests_cache.CachedSession",
        return_value=Session(),
    )

    url = "https://api.github.com/repos/apache/superset/pulls?state=all&per_page=100&page=1"
    requests_mock.get(
        url,
        status_code=403,
        json={
            "message": (
                "API rate limit exceeded for 66.220.13.38. (But here's the good "
                "news: Authenticated requests get a higher rate limit. Check out "
                "the documentation for more details.)"
            ),
            "documentation_url": (
                "https://docs.github.com/rest/overview/"
                "resources-in-the-rest-api#rate-limiting"
            ),
        },
    )

    connection = connect(":memory:")
    cursor = connection.cursor()

    sql = """
        SELECT * FROM
        "https://api.github.com/repos/apache/superset/pulls"
    """
    with pytest.raises(ProgrammingError) as excinfo:
        list(cursor.execute(sql))
    assert str(excinfo.value) == (
        "API rate limit exceeded for 66.220.13.38. (But here's the good "
        "news: Authenticated requests get a higher rate limit. Check out "
        "the documentation for more details.)"
    )


def test_github_auth_token(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test a simple request.
    """
    mocker.patch(
        "shillelagh.adapters.api.github.requests_cache.CachedSession",
        return_value=Session(),
    )

    single_resource_url = "https://api.github.com/repos/apache/superset/pulls/16581"
    single_resource_mock = requests_mock.get(
        single_resource_url,
        json=github_single_response,
    )
    multiple_resources_url = (
        "https://api.github.com/repos/apache/superset/pulls?"
        "state=all&per_page=100&page=1"
    )
    multiple_resources_mock = requests_mock.get(multiple_resources_url, json=[])

    connection = connect(
        ":memory:",
        adapter_kwargs={"githubapi": {"access_token": "XXX"}},
    )
    cursor = connection.cursor()

    sql = """
        SELECT * FROM
        "https://api.github.com/repos/apache/superset/pulls"
        WHERE number = 16581
    """
    list(cursor.execute(sql))
    assert single_resource_mock.last_request.headers["Authorization"] == "Bearer XXX"

    sql = """
        SELECT * FROM
        "https://api.github.com/repos/apache/superset/pulls"
    """
    list(cursor.execute(sql))
    assert multiple_resources_mock.last_request.headers["Authorization"] == "Bearer XXX"



"""
Test number parsing.
"""
import pytest

from shillelagh.adapters.api.gsheets.parsing.base import tokenize
from shillelagh.adapters.api.gsheets.parsing.number import (
    AT,
    COLOR,
    COMMA,
    DIGITS,
    FRACTION,
    LITERAL,
    PERCENT,
    PERIOD,
    STAR,
    UNDERSCORE,
    E,
    InvalidValue,
    condition_matches,
    fix_periods,
    format_number_pattern,
    get_fraction,
    has_condition,
    parse_number_pattern,
)


def test_digits_token():
    """
    Test the digits token.
    """
    token = DIGITS("00")
    assert token.format(123, [token]) == "123"
    operation, rest = token.parse("123", [token])
    assert operation["operation"](0) == 123
    assert rest == ""

    token = DIGITS("0000")
    assert token.format(123, [token]) == "0123"
    operation, rest = token.parse("0123", [token])
    assert operation["operation"](0) == 123
    assert rest == ""

    token = DIGITS("00")
    assert token.format(3, [token]) == "03"
    operation, rest = token.parse("03", [token])
    assert operation["operation"](0) == 3
    assert rest == ""

    token = DIGITS("#0")
    assert token.format(3, [token]) == "3"
    operation, rest = token.parse("3", [token])
    assert operation["operation"](0) == 3
    assert rest == ""

    token = DIGITS("?0")
    assert token.format(3, [token]) == " 3"
    operation, rest = token.parse("3", [token])
    assert operation["operation"](0) == 3
    assert rest == ""

    token = DIGITS("00")
    assert token.format(0.456, [DIGITS("0"), PERIOD("."), token]) == "45"
    operation, rest = token.parse("45", [DIGITS("0"), PERIOD("."), token])
    assert operation["operation"](0) == 0.45
    assert rest == ""

    token = DIGITS("#,###")
    assert token.format(12000, [token]) == "12,000"
    operation, rest = token.parse("12,000", [token])
    assert operation["operation"](0) == 12000
    assert rest == ""

    token = DIGITS("#,##0")
    assert token.format(1234513400, [token]) == "1,234,513,400"
    operation, rest = token.parse("1,234,513,400", [token])
    assert operation["operation"](0) == 1234513400
    assert rest == ""

    token = DIGITS("???")
    operation, rest = token.parse("  1", [token])
    assert operation["operation"](0) == 1
    assert rest == ""


def test_digits_errors():
    """
    Test errors with the digits token.
    """
    token = DIGITS("00")
    with pytest.raises(InvalidValue) as excinfo:
        token.parse("invalid", [token])
    assert str(excinfo.value) == "invalid"

    token = DIGITS("invalid")
    with pytest.raises(Exception) as excinfo:
        token.format(123, [token])
    assert str(excinfo.value) == "Invalid token: d"

    token = DIGITS("invalid")
    with pytest.raises(Exception) as excinfo:
        token.format(0.456, [DIGITS("0"), PERIOD("."), token])
    assert str(excinfo.value) == "Invalid token: i"


def test_period_token():
    """
    Test the period token.
    """
    token = PERIOD(".")
    assert token.format(123, [token]) == "."
    operation, rest = token.parse(".123", [token])
    assert operation == {}
    assert rest == "123"

    with pytest.raises(InvalidValue) as excinfo:
        token.parse("invalid", [token])
    assert str(excinfo.value) == "invalid"


def test_multiple_periods():
    """
    Test that only the first period is tokenized to a ``PERIOD``.
    """
    tokens = list(fix_periods(tokenize("#.#...", [DIGITS, PERIOD, LITERAL])))
    assert tokens == [
        DIGITS("#"),
        PERIOD("."),
        DIGITS("#"),
        LITERAL("."),
        LITERAL("."),
        LITERAL("."),
    ]


def test_percent_token():
    """
    Test the percent token.
    """
    token = PERCENT("%")
    assert token.format(123, [token]) == "%"
    operation, rest = token.parse("%", [token])
    assert operation["operation"](80) == 0.8
    assert rest == ""

    with pytest.raises(InvalidValue) as excinfo:
        token.parse("invalid", [token])
    assert str(excinfo.value) == "invalid"


def test_comma_token():
    """
    Test the comma token.
    """
    token = COMMA(",,")
    assert token.format(123456789, [token]) == ""
    operation, rest = token.parse("", [token])
    assert operation["operation"](1.2) == 1200000
    assert rest == ""


def test_e_token():
    """
    Test the scientific notation token.
    """
    token = E("e+00")
    assert token.format(123456789, [token]) == "e+08"
    operation, rest = token.parse("e+08", [token])
    assert operation["operation"](1.2) == 120000000.0
    assert rest == ""

    assert token.format(0.123456789, [token]) == "e-01"
    operation, rest = token.parse("e-01", [token])
    assert operation["operation"](1.2) == 0.12
    assert rest == ""

    token = E("E-##")
    assert token.format(123456789, [token]) == "E8"
    operation, rest = token.parse("E8", [token])
    assert operation["operation"](1.2) == 120000000.0
    assert rest == ""

    assert token.format(0.123456789, [token]) == "E-1"
    operation, rest = token.parse("E-1", [token])
    assert operation["operation"](1.2) == 0.12
    assert rest == ""

    token = E("invalid")
    with pytest.raises(Exception) as excinfo:
        token.format(123, [token])
    assert str(excinfo.value) == "You are likely to be eaten by a grue."


def test_fraction_token():
    """
    Test the fraction token.
    """
    token = FRACTION("#/#")
    assert token.format(5.25, [token]) == "1/4"
    operation, rest = token.parse("1/4", [token])
    assert operation["operation"](5) == 5.25
    assert rest == ""

    token = FRACTION("#/8")
    assert token.format(5.25, [token]) == "2/8"
    assert operation["operation"](5) == 5.25
    assert rest == ""

    token = FRACTION("???/???")
    operation, rest = token.parse("  1/8", [token])
    assert operation["operation"](5) == 5.125
    assert rest == ""

    token = FRACTION("invalid")
    with pytest.raises(Exception) as excinfo:
        token.format(123, [token])
    assert str(excinfo.value) == "You are likely to be eaten by a grue."


def test_star_token():
    """
    Test the star token.
    """
    token = STAR("*")
    assert token.format(123, [token]) == ""
    operation, rest = token.parse("1", [token])
    assert operation == {}
    assert rest == "1"


def test_underscore_token():
    """
    Test the underscore token.
    """
    token = UNDERSCORE("_)")
    assert token.format(123, [token]) == " "
    operation, rest = token.parse(" ", [token])
    assert operation == {}
    assert rest == ""

    token = UNDERSCORE("_)")
    with pytest.raises(Exception) as excinfo:
        token.parse("A", [token])
    assert str(excinfo.value) == "A"


def test_at_token():
    """
    Test the at token.
    """
    token = AT("@")
    assert token.format("123", [token]) == "123"
    operation, rest = token.parse("123", [token])
    assert operation["operation"](0) == "123"
    assert rest == ""

    # test lookahead
    token = AT("@")
    operation, rest = token.parse("123 ", [token, UNDERSCORE("_)")])
    assert operation["operation"](0) == "123"
    assert rest == " "

    with pytest.raises(InvalidValue) as excinfo:
        token.parse("123$", [token, UNDERSCORE("_)")])
    assert str(excinfo.value) == "123$"


def test_color_token():
    """
    Test the color token.
    """
    token = COLOR("[Black]")
    assert token.format(123, [token]) == ""
    operation, rest = token.parse("123", [token])
    assert operation == {}
    assert rest == "123"


def test_get_fraction() -> None:
    """
    Test ``get_fraction``.
    """
    assert get_fraction(9) == 0.9
    assert get_fraction(123) == 0.123
    assert get_fraction(1) == 0.1
    assert get_fraction(10) == 0.1
    assert get_fraction(0) == 0

    with pytest.raises(Exception) as excinfo:
        get_fraction(-1)
    assert str(excinfo.value) == "Number should be a positive integer"


def test_has_condition() -> None:
    """
    Test ``has_condition``.
    """
    assert has_condition('###0.000;"TEXT: "_(@_)') is False
    assert has_condition("[Blue]#,##0;[Red]#,##0;[Green]0.0;[Magenta]_(@_)") is False
    assert has_condition('[>1000]"HIGH";[Color43][<=200]"LOW";0000') is True


def test_condition_matches() -> None:
    """
    Test ``condition_matches``.
    """
    assert condition_matches(1005, '[>1000]"HIGH"') is True
    assert condition_matches(32, '[>1000]"HIGH"') is False
    assert condition_matches(527, '[>1000]"HIGH"') is False

    assert condition_matches(1005, '[Color43][<=200]"LOW"') is False
    assert condition_matches(32, '[Color43][<=200]"LOW"') is True
    assert condition_matches(527, '[Color43][<=200]"LOW"') is False

    assert condition_matches(1005, "0000") is True
    assert condition_matches(32, "0000") is True
    assert condition_matches(527, "0000") is True


def test_parse_number_pattern() -> None:
    """
    Test ``parse_number_pattern``.
    """
    # examples from the spec
    assert parse_number_pattern("12345.1", "####.#") == 12345.1
    assert parse_number_pattern("012.3400", "000.0000") == 12.34
    assert parse_number_pattern("12.0", "#.0#") == 12
    assert parse_number_pattern("5   1/8  ", "# ???/???") == 5.125
    assert parse_number_pattern("23 1/4", "0 #/#") == 23.25
    assert parse_number_pattern("23 2/8", "0 #/8") == 23.25
    assert parse_number_pattern("23 ", "0 #/3") == 23
    assert parse_number_pattern("12,000", "#,###") == 12000
    assert parse_number_pattern("1.2M", '0.0,,"M"') == 1200000
    assert parse_number_pattern("1.23e+09", "0.00e+00") == 1230000000
    assert parse_number_pattern("1,234,513,400.00", "#,##0.00") == 1234513400.0

    # additional examples
    assert parse_number_pattern("-12345.1", "####.#") == -12345.1
    assert parse_number_pattern("80%", "##%") == 0.8
    assert parse_number_pattern("80.00", "#,##0.00") == 80.0
    assert parse_number_pattern("12.2", "#0.0,,") == 12200000
    assert parse_number_pattern("12.2", "*#0.0,,") == 12200000
    assert parse_number_pattern("$12.2", "$#0.0,,") == 12200000
    assert parse_number_pattern("12.2,", r"#0.0\,") == 12.2
    assert parse_number_pattern("dollars: 12.0", "dollars: #.0#") == 12
    assert parse_number_pattern("1.23e-02", "0.00e+00") == 0.0123
    assert parse_number_pattern("0123.45", "#0000.00") == 123.45

    # tests for conditions
    assert (
        parse_number_pattern("0527", '[>1000]"HIGH";[Color43][<=200]"LOW";0000') == 527
    )

    # some really complicated examples
    assert parse_number_pattern("123.00 ", "#,##0.00_);[Red](#,##0.00)") == 123
    assert parse_number_pattern("(123.00)", "#,##0.00_);[Red](#,##0.00)") == -123
    assert parse_number_pattern("TEXT:  MyText ", '###0.000;"TEXT: "_(@_)') == "MyText"
    assert (
        parse_number_pattern(
            " $ 123.00 ",
            '_($* #,##0.00_);_($* (#,##0.00);_($* "-"??_);_(@_)',
        )
        == 123
    )


def test_parse_number_pattern_errors() -> None:
    """
    Test errors in ``parse_number_pattern``.
    """
    with pytest.raises(Exception) as excinfo:
        parse_number_pattern("80", "##%")
    assert str(excinfo.value) == "Unable to parse value 80 with pattern ##%"

    with pytest.raises(Exception) as excinfo:
        parse_number_pattern("1.23", "0.00e+00")
    assert str(excinfo.value) == "Unable to parse value 1.23 with pattern 0.00e+00"


def test_format_number_pattern() -> None:
    """
    Test ``format_number_pattern``.
    """
    # examples from the spec
    assert format_number_pattern(12345.1, "####.#") == "12345.1"
    assert format_number_pattern(12.34, "000.0000") == "012.3400"
    assert format_number_pattern(12, "#.0#") == "12.0"
    assert format_number_pattern(5.125, "# ???/???") == "5   1/8  "
    assert format_number_pattern(23.25, "0 #/#") == "23 1/4"
    assert format_number_pattern(23.25, "0 #/8") == "23 2/8"
    assert format_number_pattern(23.1, "0 #/3") == "23 "
    assert format_number_pattern(12000, "#,###") == "12,000"
    assert format_number_pattern(1200000, '0.0,,"M"') == "1.2M"
    assert format_number_pattern(1230000000, "0.00e+00") == "1.23e+09"
    assert format_number_pattern(1234513400.0, "#,##0.00") == "1,234,513,400.00"

    # additional examples
    assert format_number_pattern(-12345.1, "####.#") == "-12345.1"
    assert format_number_pattern(0.8, "##%") == "80%"
    assert format_number_pattern(80.0, "#,##0.00") == "80.00"
    assert format_number_pattern(12200000, "#0.0,,") == "12.2"
    assert format_number_pattern(12200000, "*#0.0,,") == "12.2"
    assert format_number_pattern(12200000, "$#0.0,,") == "$12.2"
    assert format_number_pattern(12.2, r"#0.0\,") == "12.2,"
    assert format_number_pattern(12, "dollars: #0.0#") == "dollars: 12.0"
    assert format_number_pattern(0.0123, "0.00e+00") == "1.23e-02"
    assert format_number_pattern(123.456, "#0000.00") == "0123.45"

    # tests for conditions
    assert (
        format_number_pattern(1005, '[>1000]"HIGH";[Color43][<=200]"LOW";0000')
        == "HIGH"
    )
    assert (
        format_number_pattern(32, '[>1000]"HIGH";[Color43][<=200]"LOW";0000') == "LOW"
    )
    assert (
        format_number_pattern(527, '[>1000]"HIGH";[Color43][<=200]"LOW";0000') == "0527"
    )
    assert (
        format_number_pattern(527, '[>1000]"HIGH";[Color43][<=200]"LOW"') == "########"
    )

    # some really complicated examples
    assert format_number_pattern(123, "#,##0.00_);[Red](#,##0.00)") == "123.00 "
    assert format_number_pattern(-123, "#,##0.00_);[Red](#,##0.00)") == "(123.00)"
    assert format_number_pattern("MyText", '###0.000;"TEXT: "_(@_)') == "TEXT:  MyText "
    assert format_number_pattern(0, "#;(#);ZERO") == "ZERO"


def test_format_number_pattern_errors() -> None:
    """
    Test errors in ``format_number_pattern``.
    """
    with pytest.raises(Exception) as excinfo:
        format_number_pattern(123, "")
    assert str(excinfo.value) == "Empty pattern!"

    with pytest.raises(Exception) as excinfo:
        format_number_pattern("MyText", "###0.000")
    assert str(excinfo.value) == "No text format found for string value"



"""
Test for shillelagh.adapter.base.
"""
from datetime import datetime
from typing import List

import pytest
from pytest_mock import MockerFixture

from shillelagh.adapters.base import Adapter
from shillelagh.backends.apsw.db import connect
from shillelagh.exceptions import NotSupportedError
from shillelagh.fields import DateTime, Order
from shillelagh.filters import Equal, Range
from shillelagh.typing import Row

from ..fakes import FakeAdapter, FakeEntryPoint


class FakeAdapterWithDateTime(FakeAdapter):

    """
    An adapter with a timestamp column.
    """

    birthday = DateTime(filters=[Range], order=Order.ANY, exact=True)

    data: List[Row] = []

    def __init__(self):  # pylint: disable=super-init-not-called
        pass


class ReadOnlyAdapter(Adapter):  # pylint: disable=abstract-method
    """
    A read-only adapter.
    """


def test_adapter_get_columns() -> None:
    """
    Test ``get_columns``.
    """
    adapter = FakeAdapter()
    assert adapter.get_columns() == {
        "age": FakeAdapter.age,
        "name": FakeAdapter.name,
        "pets": FakeAdapter.pets,
    }
    adapter.close()


def test_adapter_get_metadata() -> None:
    """
    Test ``get_metadata``.
    """
    adapter = FakeAdapter()
    assert adapter.get_metadata() == {}


def test_adapter_read_only() -> None:
    """
    Test a read-only adapter.
    """
    adapter = ReadOnlyAdapter()

    with pytest.raises(NotSupportedError) as excinfo:
        adapter.insert_data({"hello": "world"})
    assert str(excinfo.value) == "Adapter does not support ``INSERT`` statements"

    with pytest.raises(NotSupportedError) as excinfo:
        adapter.delete_data(1)
    assert str(excinfo.value) == "Adapter does not support ``DELETE`` statements"

    with pytest.raises(NotSupportedError) as excinfo:
        adapter.update_data(1, {"hello": "universe"})
    assert str(excinfo.value) == "Adapter does not support ``DELETE`` statements"


def test_adapter_get_data() -> None:
    """
    Test ``get_data``.
    """
    adapter = FakeAdapter()

    data = adapter.get_data({}, [])
    assert list(data) == [
        {"rowid": 0, "name": "Alice", "age": 20, "pets": 0},
        {"rowid": 1, "name": "Bob", "age": 23, "pets": 3},
    ]

    data = adapter.get_data({"name": Equal("Alice")}, [])
    assert list(data) == [
        {"rowid": 0, "name": "Alice", "age": 20, "pets": 0},
    ]

    data = adapter.get_data({"age": Range(20, None, False, False)}, [])
    assert list(data) == [
        {"rowid": 1, "name": "Bob", "age": 23, "pets": 3},
    ]

    data = adapter.get_data({}, [("age", Order.DESCENDING)])
    assert list(data) == [
        {"rowid": 1, "name": "Bob", "age": 23, "pets": 3},
        {"rowid": 0, "name": "Alice", "age": 20, "pets": 0},
    ]


def test_adapter_get_rows() -> None:
    """
    Test ``get_rows``.
    """
    adapter = FakeAdapter()

    adapter.insert_row({"rowid": None, "name": "Charlie", "age": 6, "pets": 1})

    data = adapter.get_data({}, [])
    assert list(data) == [
        {"rowid": 0, "name": "Alice", "age": 20, "pets": 0},
        {"rowid": 1, "name": "Bob", "age": 23, "pets": 3},
        {"rowid": 2, "name": "Charlie", "age": 6.0, "pets": 1},
    ]

    data = adapter.get_rows({}, [])
    assert list(data) == [
        {"rowid": 0, "name": "Alice", "age": 20.0, "pets": 0},
        {"rowid": 1, "name": "Bob", "age": 23.0, "pets": 3},
        {"rowid": 2, "name": "Charlie", "age": 6.0, "pets": 1},
    ]


def test_adapter_manipulate_rows() -> None:
    """
    Test ``DML``.
    """
    adapter = FakeAdapter()

    adapter.insert_row({"rowid": None, "name": "Charlie", "age": 6, "pets": 1})
    data = adapter.get_data({}, [])
    assert list(data) == [
        {"rowid": 0, "name": "Alice", "age": 20, "pets": 0},
        {"rowid": 1, "name": "Bob", "age": 23, "pets": 3},
        {"rowid": 2, "name": "Charlie", "age": 6, "pets": 1},
    ]
    adapter.insert_row({"rowid": 4, "name": "Dani", "age": 40, "pets": 2})
    data = adapter.get_data({}, [])
    assert list(data) == [
        {"rowid": 0, "name": "Alice", "age": 20, "pets": 0},
        {"rowid": 1, "name": "Bob", "age": 23, "pets": 3},
        {"rowid": 2, "name": "Charlie", "age": 6, "pets": 1},
        {"rowid": 4, "name": "Dani", "age": 40, "pets": 2},
    ]

    adapter.delete_row(0)
    data = adapter.get_data({}, [])
    assert list(data) == [
        {"rowid": 1, "name": "Bob", "age": 23, "pets": 3},
        {"rowid": 2, "name": "Charlie", "age": 6, "pets": 1},
        {"rowid": 4, "name": "Dani", "age": 40, "pets": 2},
    ]

    adapter.update_row(1, {"rowid": 1, "name": "Bob", "age": 24, "pets": 4})
    data = adapter.get_data({}, [])
    assert list(data) == [
        {"rowid": 2, "name": "Charlie", "age": 6, "pets": 1},
        {"rowid": 4, "name": "Dani", "age": 40, "pets": 2},
        {"rowid": 1, "name": "Bob", "age": 24, "pets": 4},
    ]


def test_type_conversion(mocker: MockerFixture) -> None:
    """
    Test that native types are converted correctly.
    """
    entry_points = [FakeEntryPoint("dummy", FakeAdapterWithDateTime)]
    mocker.patch(
        "shillelagh.backends.apsw.db.iter_entry_points",
        return_value=entry_points,
    )

    connection = connect(":memory:", ["dummy"], isolation_level="IMMEDIATE")
    cursor = connection.cursor()

    cursor.execute('SELECT * FROM "dummy://"')
    assert cursor.fetchall() == []

    cursor.execute(
        'INSERT INTO "dummy://" (birthday) VALUES (?)',
        (datetime(2021, 1, 1, 0, 0),),
    )
    cursor.execute('SELECT * FROM "dummy://"')
    assert cursor.fetchall() == [
        (
            None,
            datetime(2021, 1, 1, 0, 0),
            None,
            None,
        ),
    ]

    # make sure datetime is stored as a datetime
    assert FakeAdapterWithDateTime.data == [
        {
            "age": None,
            "birthday": datetime(2021, 1, 1, 0, 0),
            "name": None,
            "pets": None,
            "rowid": 1,
        },
    ]
    assert isinstance(FakeAdapterWithDateTime.data[0]["birthday"], datetime)

    cursor.execute(
        'SELECT * FROM "dummy://" WHERE birthday > ?',
        (datetime(2020, 12, 31, 0, 0),),
    )
    assert cursor.fetchall() == [
        (None, datetime(2021, 1, 1, 0, 0), None, None),
    ]



# pylint: disable=invalid-name, c-extension-no-member, no-self-use, unused-import
"""
A DB API 2.0 wrapper for APSW.
"""
import datetime
import itertools
import logging
from collections import Counter
from functools import partial, wraps
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    cast,
)

import apsw
from pkg_resources import iter_entry_points

from shillelagh import functions
from shillelagh.adapters.base import Adapter
from shillelagh.backends.apsw.vt import VTModule, type_map
from shillelagh.exceptions import Warning  # pylint: disable=redefined-builtin
from shillelagh.exceptions import (
    DatabaseError,
    DataError,
    Error,
    IntegrityError,
    InterfaceError,
    InternalError,
    NotSupportedError,
    OperationalError,
    ProgrammingError,
)
from shillelagh.fields import Blob, Field
from shillelagh.lib import combine_args_kwargs, escape, find_adapter, serialize
from shillelagh.types import (
    BINARY,
    DATETIME,
    NUMBER,
    ROWID,
    STRING,
    Binary,
    Date,
    DateFromTicks,
    Time,
    TimeFromTicks,
    Timestamp,
    TimestampFromTicks,
)
from shillelagh.typing import Description, SQLiteValidType

apilevel = "2.0"
threadsafety = 2
paramstyle = "qmark"
sqlite_version_info = tuple(
    int(number) for number in apsw.sqlitelibversion().split(".")
)

NO_SUCH_TABLE = "SQLError: no such table: "
SCHEMA = "main"

CURSOR_METHOD = TypeVar("CURSOR_METHOD", bound=Callable[..., Any])

_logger = logging.getLogger(__name__)


def check_closed(method: CURSOR_METHOD) -> CURSOR_METHOD:
    """Decorator that checks if a connection or cursor is closed."""

    @wraps(method)
    def wrapper(self: "Cursor", *args: Any, **kwargs: Any) -> Any:
        if self.closed:
            raise ProgrammingError(f"{self.__class__.__name__} already closed")
        return method(self, *args, **kwargs)

    return cast(CURSOR_METHOD, wrapper)


def check_result(method: CURSOR_METHOD) -> CURSOR_METHOD:
    """Decorator that checks if the cursor has results from ``execute``."""

    @wraps(method)
    def wrapper(self: "Cursor", *args: Any, **kwargs: Any) -> Any:
        if self._results is None:  # pylint: disable=protected-access
            raise ProgrammingError("Called before ``execute``")
        return method(self, *args, **kwargs)

    return cast(CURSOR_METHOD, wrapper)


def get_type_code(type_name: str) -> Type[Field]:
    """
    Return a ``Field`` that corresponds to a type name.

    This is used to build the description of the cursor after a successful
    query.
    """
    return cast(Type[Field], type_map.get(type_name, Blob))


def convert_binding(binding: Any) -> SQLiteValidType:
    """
    Convert a binding to a SQLite type.

    Eg, if the user is filtering a timestamp column we need to convert the
    ``datetime.datetime`` object in the binding to a string.
    """
    if isinstance(binding, bool):
        return int(binding)
    if isinstance(binding, (int, float, str, bytes, type(None))):
        return binding
    if isinstance(binding, (datetime.datetime, datetime.date, datetime.time)):
        return binding.isoformat()
    return str(binding)


class Cursor:  # pylint: disable=too-many-instance-attributes

    """
    Connection cursor.
    """

    def __init__(
        self,
        cursor: "apsw.Cursor",
        adapters: List[Type[Adapter]],
        adapter_kwargs: Dict[str, Dict[str, Any]],
        isolation_level: Optional[str] = None,
    ):
        self._cursor = cursor
        self._adapters = adapters
        self._adapter_kwargs = adapter_kwargs

        self.in_transaction = False
        self.isolation_level = isolation_level

        # This read/write attribute specifies the number of rows to fetch at a
        # time with .fetchmany(). It defaults to 1 meaning to fetch a single
        # row at a time.
        self.arraysize = 1

        self.closed = False

        # this is updated only after a query
        self.description: Description = None

        # this is set to an iterator of rows after a successful query
        self._results: Optional[Iterator[Tuple[Any, ...]]] = None
        self._rowcount = -1

    @property  # type: ignore
    @check_closed
    def rowcount(self) -> int:
        """
        Return the number of rows after a query.
        """
        try:
            results = list(self._results)  # type: ignore
        except TypeError:
            return -1

        n = len(results)
        self._results = iter(results)
        return max(0, self._rowcount) + n

    @check_closed
    def close(self) -> None:
        """
        Close the cursor.
        """
        self._cursor.close()
        self.closed = True

    @check_closed
    def execute(
        self,
        operation: str,
        parameters: Optional[Tuple[Any, ...]] = None,
    ) -> "Cursor":
        """
        Execute a query using the cursor.
        """
        if not self.in_transaction and self.isolation_level:
            self._cursor.execute(f"BEGIN {self.isolation_level}")
            self.in_transaction = True

        self.description = None
        self._rowcount = -1

        # convert parameters (bindings) to types accepted by SQLite
        if parameters:
            parameters = tuple(convert_binding(parameter) for parameter in parameters)

        # this is where the magic happens: instead of forcing users to register
        # their virtual tables explicitly, we do it for them when they first try
        # to access them and it fails because the table doesn't exist yet
        while True:
            try:
                self._cursor.execute(operation, parameters)
                self.description = self._get_description()
                self._results = self._convert(self._cursor)
                break
            except apsw.SQLError as ex:
                message = ex.args[0]
                if not message.startswith(NO_SUCH_TABLE):
                    raise ProgrammingError(message) from ex

                # create the virtual table
                uri = message[len(NO_SUCH_TABLE) :]
                self._create_table(uri)

        return self

    def _convert(self, cursor: "apsw.Cursor") -> Iterator[Tuple[Any, ...]]:
        """
        Convert row from SQLite types to native Python types.

        SQLite only supports 5 types. For booleans and time-related types
        we need to do the conversion here.
        """
        if not self.description:
            return

        for row in cursor:
            yield tuple(
                # convert from SQLite types to native Python types
                type_map[desc[1].type]().parse(col)
                for col, desc in zip(row, self.description)
            )

    def _create_table(self, uri: str) -> None:
        """
        Create a virtual table.

        This method is called the first time a virtual table is accessed.
        """
        prefix = f"{SCHEMA}."
        if uri.startswith(prefix):
            uri = uri[len(prefix) :]

        # collect arguments from URI and connection and serialize them
        adapter = find_adapter(uri, self._adapter_kwargs, self._adapters)
        key = adapter.__name__.lower()
        args = adapter.parse_uri(uri)
        kwargs = self._adapter_kwargs.get(key, {})
        formatted_args = ", ".join(
            serialize(arg) for arg in combine_args_kwargs(adapter, *args, **kwargs)
        )
        table_name = escape(uri)
        self._cursor.execute(
            f'CREATE VIRTUAL TABLE "{table_name}" USING {adapter.__name__}({formatted_args})',
        )

    def _get_description(self) -> Description:
        """
        Return the cursor description.

        We only return name and type, since that's what we get from APSW.
        """
        try:
            description = self._cursor.getdescription()
        except apsw.ExecutionCompleteError:
            return None

        return [
            (
                name,
                get_type_code(type_name),
                None,
                None,
                None,
                None,
                True,
            )
            for name, type_name in description
        ]

    @check_closed
    def executemany(
        self,
        operation: str,
        seq_of_parameters: Optional[List[Tuple[Any, ...]]] = None,
    ) -> "Cursor":
        """
        Execute multiple statements.

        Currently not supported.
        """
        raise NotSupportedError(
            "``executemany`` is not supported, use ``execute`` instead",
        )

    @check_result
    @check_closed
    def fetchone(self) -> Optional[Tuple[Any, ...]]:
        """
        Fetch the next row of a query result set, returning a single sequence,
        or ``None`` when no more data is available.
        """
        try:
            row = self.next()
        except StopIteration:
            return None

        self._rowcount = max(0, self._rowcount) + 1

        return row

    @check_result
    @check_closed
    def fetchmany(self, size=None) -> List[Tuple[Any, ...]]:
        """
        Fetch the next set of rows of a query result, returning a sequence of
        sequences (e.g. a list of tuples). An empty sequence is returned when
        no more rows are available.
        """
        size = size or self.arraysize
        results = list(itertools.islice(self, size))

        return results

    @check_result
    @check_closed
    def fetchall(self) -> List[Tuple[Any, ...]]:
        """
        Fetch all (remaining) rows of a query result, returning them as a
        sequence of sequences (e.g. a list of tuples). Note that the cursor's
        arraysize attribute can affect the performance of this operation.
        """
        results = list(self)

        return results

    @check_closed
    def setinputsizes(self, sizes: int) -> None:
        """
        Used before ``execute`` to predefine memory areas for parameters.

        Currently not supported.
        """

    @check_closed
    def setoutputsizes(self, sizes: int) -> None:
        """
        Set a column buffer size for fetches of large columns.

        Currently not supported.
        """

    @check_result
    @check_closed
    def __iter__(self) -> Iterator[Tuple[Any, ...]]:
        for row in self._results:  # type: ignore
            self._rowcount = max(0, self._rowcount) + 1
            yield row

    @check_result
    @check_closed
    def __next__(self) -> Tuple[Any, ...]:
        return next(self._results)  # type: ignore

    next = __next__


def apsw_version() -> str:
    """
    Custom implementation of the ``VERSION`` function.

    This function shows the backend version::

        sql> SELECT VERSION();
        VERSION()
        ----------------------
        1.0.5 (apsw 3.36.0-r1)

    """
    return f"{functions.version()} (apsw {apsw.apswversion()})"


class Connection:

    """Connection."""

    def __init__(
        self,
        path: str,
        adapters: List[Type[Adapter]],
        adapter_kwargs: Dict[str, Dict[str, Any]],
        isolation_level: Optional[str] = None,
    ):
        # create underlying APSW connection
        self._connection = apsw.Connection(path)
        self.isolation_level = isolation_level

        # register adapters
        for adapter in adapters:
            self._connection.createmodule(adapter.__name__, VTModule(adapter))
        self._adapters = adapters
        self._adapter_kwargs = adapter_kwargs

        # register functions
        available_functions = {
            "sleep": functions.sleep,
            "version": apsw_version,
            "get_metadata": partial(
                functions.get_metadata,
                self._adapter_kwargs,
                adapters,
            ),
        }
        for name, function in available_functions.items():
            self._connection.createscalarfunction(name, function)

        self.closed = False
        self.cursors: List[Cursor] = []

    @check_closed
    def close(self) -> None:
        """Close the connection now."""
        self.closed = True
        for cursor in self.cursors:
            if not cursor.closed:
                cursor.close()

    @check_closed
    def commit(self) -> None:
        """Commit any pending transaction to the database."""
        for cursor in self.cursors:
            if cursor.in_transaction:
                cursor._cursor.execute("COMMIT")  # pylint: disable=protected-access
                cursor.in_transaction = False

    @check_closed
    def rollback(self) -> None:
        """Rollback any transactions."""
        for cursor in self.cursors:
            if cursor.in_transaction:
                cursor._cursor.execute("ROLLBACK")  # pylint: disable=protected-access
                cursor.in_transaction = False

    @check_closed
    def cursor(self) -> Cursor:
        """Return a new Cursor Object using the connection."""
        cursor = Cursor(
            self._connection.cursor(),
            self._adapters,
            self._adapter_kwargs,
            self.isolation_level,
        )
        self.cursors.append(cursor)

        return cursor

    @check_closed
    def execute(
        self,
        operation: str,
        parameters: Optional[Tuple[Any, ...]] = None,
    ) -> Cursor:
        """
        Execute a query on a cursor.
        """
        cursor = self.cursor()
        return cursor.execute(operation, parameters)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.commit()
        self.close()


def connect(
    path: str,
    adapters: Optional[List[str]] = None,
    adapter_kwargs: Optional[Dict[str, Dict[str, Any]]] = None,
    safe: bool = False,
    isolation_level: Optional[str] = None,
) -> Connection:
    r"""
    Constructor for creating a connection to the database.

        >>> conn = connect("database.sqlite", ["csvfile", "weatherapi"])
        >>> curs = conn.cursor()

    Let's create a fake CSV file and access it:

        >>> from pyfakefs.fake_filesystem_unittest import Patcher
        >>> with Patcher() as patcher:
        ...     fake_file = patcher.fs.create_file('/foo/bar.csv', contents='"a","b"\n1,2\n3,4')
        ...     list(curs.execute("SELECT * FROM '/foo/bar.csv'"))
        [(1.0, 2.0), (3.0, 4.0)]

    """
    if adapters is None and safe:
        adapters = []
    adapter_kwargs = adapter_kwargs or {}

    all_adapters = []
    for entry_point in iter_entry_points("shillelagh.adapter"):
        if adapters is None or entry_point.name in adapters:
            try:
                adapter = entry_point.load()
            except (ImportError, ModuleNotFoundError) as ex:
                _logger.warning("Couldn't load adapter %s", entry_point.name)
                _logger.debug(ex)
                continue
            all_adapters.append((entry_point.name, adapter))

    all_adapters_names = [name for name, adapter in all_adapters]

    # check if there are any repeated names, to prevent malicious adapters
    if safe:
        repeated = {
            name for name, count in Counter(all_adapters_names).items() if count > 1
        }
        if repeated:
            raise InterfaceError(f'Repeated adapter names found: {", ".join(repeated)}')

    if adapters is None:
        adapters = all_adapters_names
    enabled_adapters = [
        adapter
        for (name, adapter) in all_adapters
        if name in adapters and (adapter.safe or not safe)
    ]

    # replace entry point names with class names
    mapping = {
        name: adapter.__name__.lower()
        for name, adapter in all_adapters
        if adapter in enabled_adapters
    }
    adapter_kwargs = {mapping[k]: v for k, v in adapter_kwargs.items()}

    return Connection(
        path,
        enabled_adapters,
        adapter_kwargs,
        isolation_level,
    )
